"""Agora WebSocket handler for Mammotion WebRTC streaming."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from dataclasses import dataclass
from http.model.camera_stream import StreamSubscriptionResponse
import ipaddress
import json
import logging
import secrets
import ssl
import time
from typing import Any

import aiohttp
from sdp_transform import parse as sdp_parse
from webrtc_models import RTCIceCandidateInit
from websockets.asyncio.client import ClientConnection, connect
from websockets.exceptions import WebSocketException

from .agora_api import AgoraResponse
from .agora_sdp import parse_offer_to_ortc

_LOGGER = logging.getLogger(__name__)


def _create_ws_ssl_context() -> ssl.SSLContext:
    """Create SSL context for WebSocket connections."""
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE
    return ssl_context


_SSL_CONTEXT = _create_ws_ssl_context()


@dataclass
class AddressEntry:
    """Agora edge server address entry."""

    ip: str
    port: int
    ticket: str


@dataclass
class ResponseInfo:
    """Agora API response information."""

    code: int
    addresses: list[AddressEntry]
    server_ts: int
    uid: int
    cid: int
    cname: str
    detail: dict[str, str]
    flag: int
    opid: int
    cert: str


@dataclass
class SdpInfo:
    """SDP parsing information."""

    parsed_sdp: dict
    fingerprint: str
    ice_ufrag: str
    ice_pwd: str
    audio_codecs: list[dict[str, Any]]
    video_codecs: list[dict[str, Any]]
    audio_extensions: list[dict[str, Any]]
    video_extensions: list[dict[str, Any]]
    audio_direction: str
    video_direction: str
    ice_candidates: list[dict[str, Any]]
    extmap_allow_mixed: bool
    setup_role: str


class AgoraWebSocketHandler:
    """Handle Agora WebSocket communications for WebRTC streaming."""

    def __init__(self) -> None:
        """Initialize the Agora WebSocket handler."""
        self._websocket: ClientConnection | None = None
        self._connection_state = "DISCONNECTED"
        self._message_handlers: dict[str, Callable] = {}
        self._response_handlers: dict[str, asyncio.Future] = {}
        self.candidates: list[RTCIceCandidateInit] = []
        self._online_users: set[int] = set()
        self._video_streams: dict[int, dict[str, Any]] = {}
        self._answer_sdp: str | None = None
        # Background tasks
        self._message_loop_task: asyncio.Task | None = None
        self._ping_task: asyncio.Task | None = None
        # Token refresh state
        self._rejoin_token: str | None = None
        self._session_id: str | None = None
        self._channel_name: str | None = None
        self._cid: int = 0
        self._uid: int = 0
        self._vid: int = 0
        self._agora_data: StreamSubscriptionResponse | None = None
        self._setup_message_handlers()

    def _setup_message_handlers(self) -> None:
        """Set up message handlers for different WebSocket message types."""
        self._message_handlers = {
            "answer": self._handle_answer,
            "on_p2p_lost": self._handle_p2p_lost,
            "error": self._handle_error,
            "on_rtp_capability_change": self._handle_rtp_capability_change,
            "on_user_online": self._handle_user_online,
            "on_user_offline": self._handle_user_offline,
            "on_add_video_stream": self._handle_add_video_stream,
        }

    async def connect_and_join(
        self,
        agora_data: StreamSubscriptionResponse,
        offer_sdp: str,
        session_id: str,
        agora_response: AgoraResponse,
    ) -> str | None:
        """Connect to Agora WebSocket and perform join negotiation.

        The WebSocket connection stays open after returning the answer SDP.
        A background task continues processing messages (subscribe, token refresh, etc.).
        Call disconnect() to close the connection.

        Note: ICE candidates should be set in self.candidates before calling this method.
        These candidates will be incorporated into the offer SDP before sending to Agora.

        """
        _LOGGER.debug("Starting Agora WebSocket connection for session %s", session_id)
        _LOGGER.info("Agora data: %s", agora_data)

        # Store for later use in token refresh / rejoin
        self._agora_data = agora_data
        self._session_id = session_id

        # Parse offer SDP for capabilities parse_offer_to_ortc
        stored_sdp_info = self._parse_offer_sdp(offer_sdp)
        ortc_info = parse_offer_to_ortc(offer_sdp)
        if not ortc_info:
            _LOGGER.error("Failed to parse offer SDP")
            return None

        # _LOGGER.info("Offer SDP: %s", offer_sdp)
        # Try each gateway address (flag 4096) with timeout
        # Use gateway addresses specifically for WebSocket connection
        gateway_addresses = agora_response.get_gateway_addresses()
        if not gateway_addresses:
            _LOGGER.warning("No gateway addresses found, falling back to default addresses")
            gateway_addresses = agora_response.addresses

        for edge_address in gateway_addresses:
            edge_ip_dashed = edge_address.ip.replace(".", "-")
            ws_url = f"wss://{edge_ip_dashed}.edge.agora.io:{edge_address.port}"

            try:
                async with asyncio.timeout(10):  # 10 second timeout for connect
                    # Open persistent WebSocket (NOT using async with)
                    websocket = await connect(ws_url, ssl=_SSL_CONTEXT, ping_timeout=30, close_timeout=30)
                    self._websocket = websocket
                    self._connection_state = "CONNECTED"
                    _LOGGER.info("Connected to Agora WebSocket: %s", ws_url)

                    # Store SDP info for later use in trickle ICE
                    self._sdp_info = ortc_info

                    # Send join message
                    join_message = self._create_join_message(
                        agora_data,
                        offer_sdp,
                        stored_sdp_info,
                        ortc_info,
                        agora_response,
                        session_id,
                    )
                    await websocket.send(json.dumps(join_message))
                    _LOGGER.info("Sent join message to Agora %s", join_message)

                    # Wait for join response and get answer SDP
                    answer_sdp = await self._wait_for_join_response(
                        websocket, session_id, stored_sdp_info, agora_response
                    )

                    if answer_sdp:
                        # Start background message loop (keeps WS open)
                        self._message_loop_task = asyncio.ensure_future(
                            self._message_loop(websocket, session_id, stored_sdp_info, agora_response)
                        )
                        # Start ping-pong keepalive (every 3s, matching Agora SDK)
                        self._ping_task = asyncio.ensure_future(self._ping_loop())
                        return answer_sdp

                    # If join failed, close this connection and try next
                    await websocket.close()
                    self._websocket = None

            except TimeoutError:
                _LOGGER.warning("Connection timeout for edge address %s, trying next", ws_url)
                if self._websocket:
                    try:
                        await self._websocket.close()
                    except Exception:
                        pass
                    self._websocket = None
                continue
            except (WebSocketException, json.JSONDecodeError) as ex:
                _LOGGER.warning("WebSocket connection failed for %s: %s", ws_url, ex)
                if self._websocket:
                    try:
                        await self._websocket.close()
                    except Exception:
                        pass
                    self._websocket = None
                continue

        # If we get here, all connection attempts failed
        _LOGGER.error("Failed to connect to any Agora edge servers")
        self._connection_state = "DISCONNECTED"
        return None

    async def _wait_for_join_response(
        self,
        websocket: ClientConnection,
        session_id: str,
        sdp_info: SdpInfo,
        agora_response: AgoraResponse = None,
    ) -> str | None:
        """Wait for join success response and return answer SDP.

        This only processes messages until join succeeds, then returns.
        Subsequent messages are handled by _message_loop.
        """
        try:
            async with asyncio.timeout(15):  # 15s timeout for join response
                async for message in websocket:
                    try:
                        response = json.loads(message)
                        _LOGGER.info("Received Agora message: %s", response)

                        message_type = response.get("_type")
                        message_id = response.get("_id")

                        # Handle responses to requests
                        if message_id and message_id in self._response_handlers:
                            future = self._response_handlers.pop(message_id)
                            if not future.done():
                                future.set_result(response)
                            continue

                        # Handle message type handlers
                        if message_type in self._message_handlers:
                            result = await self._message_handlers[message_type](response)
                            if result:
                                return result

                        # Check for successful join response
                        if response.get("_result") == "success":
                            return await self._handle_join_success(response, sdp_info, agora_response)

                    except json.JSONDecodeError as ex:
                        _LOGGER.error("Failed to parse Agora message: %s", ex)

        except TimeoutError:
            _LOGGER.error("Timeout waiting for join response")
        except WebSocketException as ex:
            _LOGGER.error("WebSocket communication error during join: %s", ex)
            self._connection_state = "DISCONNECTED"

        # Fallback: generate basic SDP if no proper response was received
        _LOGGER.warning("No proper WebSocket response received, generating fallback SDP")
        return self._generate_fallback_sdp()

    async def _message_loop(
        self,
        websocket: ClientConnection,
        session_id: str,
        sdp_info: SdpInfo,
        agora_response: AgoraResponse = None,
    ) -> None:
        """Background message loop — stays running after join.

        Handles: on_add_video_stream, subscribe, token refresh, p2p_lost, etc.
        """
        _LOGGER.info("Started background message loop for session %s", session_id)
        try:
            async for message in websocket:
                try:
                    response = json.loads(message)
                    message_type = response.get("_type", "")
                    message_result = response.get("_result", "")
                    msg_body = response.get("_message", {})

                    # Ping response — no-op, just confirms keepalive
                    if message_result == "success" and not msg_body.get("ortc"):
                        continue

                    # Log all messages (non-ping)
                    _LOGGER.info(
                        "[msg_loop] type=%s result=%s msg=%s",
                        message_type,
                        message_result,
                        response,
                    )

                    # Dispatch to handlers
                    if message_type in self._message_handlers:
                        await self._message_handlers[message_type](response)

                    # Handle token expiry events
                    if message_type == "on_token_privilege_will_expire":
                        _LOGGER.warning("Token will expire soon, sending renew_token")
                        await self._send_renew_token()

                    elif message_type == "on_token_privilege_did_expire":
                        _LOGGER.error("Token expired! Connection may drop.")

                except json.JSONDecodeError as ex:
                    _LOGGER.error("[msg_loop] Failed to parse message: %s", ex)

        except asyncio.CancelledError:
            _LOGGER.info("Message loop cancelled")
        except WebSocketException as ex:
            _LOGGER.warning("WebSocket closed in message loop: %s", ex)
        finally:
            self._connection_state = "DISCONNECTED"
            _LOGGER.info("Message loop ended")

    async def _ping_loop(self) -> None:
        """Send ping messages every 3 seconds to keep the WebSocket alive.

        Matches the Agora SDK's handlePingPong interval.
        """
        _LOGGER.info("Started ping-pong keepalive (3s interval)")
        try:
            while self._websocket and self._connection_state == "CONNECTED":
                await asyncio.sleep(3)
                if self._websocket:
                    try:
                        ping_msg = {
                            "_id": secrets.token_hex(3),
                            "_type": "ping",
                        }
                        await self._websocket.send(json.dumps(ping_msg))
                    except (WebSocketException, ConnectionError) as ex:
                        _LOGGER.warning("Ping failed: %s", ex)
                        break
        except asyncio.CancelledError:
            _LOGGER.info("Ping loop cancelled")

    async def _send_renew_token(self) -> None:
        """Send renew_token message with current token."""
        if not self._websocket or not self._agora_data:
            return
        try:
            renew_msg = {
                "_id": secrets.token_hex(3),
                "_type": "renew_token",
                "_message": {"token": self._agora_data.token},
            }
            await self._websocket.send(json.dumps(renew_msg))
            _LOGGER.info("Sent renew_token to gateway")
        except (WebSocketException, ConnectionError) as ex:
            _LOGGER.error("Failed to send renew_token: %s", ex)

    async def _handle_join_success(
        self,
        response: dict[str, Any],
        sdp_info: SdpInfo,
        agora_response: AgoraResponse = None,
    ) -> str | None:
        """Handle successful join response and generate answer SDP."""
        message = response.get("_message", {})
        ortc = message.get("ortc", {})

        # Mark as joined so trickle ICE can start
        self._joined = True
        _LOGGER.info("Join successful - trickle ICE now enabled")

        # Store rejoin token for reconnection (mirrors agoraRTC_N.js)
        self._rejoin_token = message.get("rejoin_token")
        self._cid = message.get("cid", 0)
        self._uid = message.get("uid", 0)
        self._vid = message.get("vid", 0)
        self._channel_name = message.get("cname", "")
        if self._rejoin_token:
            _LOGGER.info("Stored rejoin_token: %s...", self._rejoin_token[:20])

        # Send set_client_role after successful connection
        # await self._send_set_client_role(role="host", level=0)

        if not ortc:
            _LOGGER.error("No ORTC parameters in join success response")
            _LOGGER.info("Full response message: %s", message)
            return None

        _LOGGER.info("ORTC parameters: %s", ortc)

        # Inject fingerprint from Agora Response (Auth) to ensure we have the correct server certificates.
        # This fixes the "bytes sent but not received" issue where DTLS fails due to missing/mismatched fingerprints.
        if agora_response:
            _LOGGER.info("Syncing fingerprints from Agora Auth Response")
            dtls_params = ortc.setdefault("dtlsParameters", {})
            current_fps = dtls_params.get("fingerprints", [])
            seen_fp_values = {f.get("fingerprint").lower() for f in current_fps if f.get("fingerprint")}

            injected_count = 0
            for addr in agora_response.addresses:
                if addr.fingerprint:
                    # Parse algorithm and value if formatted as "algo val"
                    fp_val = addr.fingerprint
                    algo = "sha-256"
                    if " " in fp_val:
                        parts = fp_val.split()
                        if len(parts) == 2:
                            algo = parts[0]
                            fp_val = parts[1]

                    if fp_val.lower() not in seen_fp_values:
                        current_fps.append({"hashFunction": algo, "fingerprint": fp_val})
                        seen_fp_values.add(fp_val.lower())
                        injected_count += 1

            if injected_count > 0:
                dtls_params["fingerprints"] = current_fps
                _LOGGER.info("Injected %d new fingerprints from Auth Response", injected_count)

        # Generate answer SDP from ORTC parameters.
        # We force 'active' role here to match Agora SDK behavior for the audience role,
        # ensuring the browser behaves as the DTLS server and Agora as the DTLS client.
        # answer_sdp = generate_answer_from_ortc(ortc, sdp_info.parsed_sdp, force_setup="active")
        answer_sdp = self._generate_answer_sdp(ortc, sdp_info)
        if answer_sdp:
            _LOGGER.info("Generated answer SDP from Agora ORTC parameters")
            # _LOGGER.info("Generated SDP: %s", answer_sdp)

            # Store answer SDP for later retrieval
            self._answer_sdp = answer_sdp

            return answer_sdp

        _LOGGER.error("Failed to generate answer SDP")
        return None

    async def _handle_answer(self, response: dict[str, Any]) -> str | None:
        """Handle answer message."""
        message = response.get("_message", {})
        sdp = message.get("sdp")
        if sdp:
            _LOGGER.info("Received direct answer SDP from Agora")
            return sdp
        return None

    async def _handle_p2p_lost(self, response: dict[str, Any]) -> None:
        """Handle P2P connection lost message."""
        error_code = response.get("error_code")
        error_str = response.get("error_str", "Unknown error")
        _LOGGER.warning("P2P connection lost: %s (code: %s)", error_str, error_code)

        # Handle specific error codes
        if error_code == 1 and "stun timeout" in error_str.lower():
            _LOGGER.info("STUN timeout detected, connection may need refreshing")
            # This could trigger a reconnection attempt

        self._connection_state = "DISCONNECTED"

    async def _handle_error(self, response: dict[str, Any]) -> None:
        """Handle error message."""
        message = response.get("_message", {})
        error = message.get("error", "Unknown error")
        _LOGGER.error("Agora WebSocket error: %s", error)

    async def _handle_rtp_capability_change(self, response: dict[str, Any]) -> None:
        """Handle RTP capability change notification."""
        message = response.get("_message", {})
        _LOGGER.info("RTP capabilities changed: %s", message)
        # Store capabilities if needed
        video_codecs = message.get("video_codec", [])
        extmap_allow_mixed = message.get("extmap_allow_mixed", False)
        web_av1_svc = message.get("web_av1_svc", False)
        _LOGGER.info(
            "Video codecs: %s, extmap_allow_mixed: %s, web_av1_svc: %s",
            video_codecs,
            extmap_allow_mixed,
            web_av1_svc,
        )

    async def _handle_user_online(self, response: dict[str, Any]) -> None:
        """Handle user online notification.

        When a user comes online, check if we already have a pending video
        stream for them (on_add_video_stream may arrive before on_user_online).
        If so, auto-subscribe now.
        """
        message = response.get("_message", {})
        uid = message.get("uid")
        if uid:
            self._online_users.add(uid)
            _LOGGER.info("User %s came online", uid)

            # Check if we already have a pending video stream for this user
            if uid in self._video_streams and self._websocket:
                stream_info = self._video_streams[uid]
                if not stream_info.get("subscribed"):
                    _LOGGER.info(
                        "User %s is online and has pending video stream, subscribing now",
                        uid,
                    )
                    stream_info["subscribed"] = True
                    await self._send_subscribe(
                        stream_id=uid,
                        ssrc_id=stream_info["ssrcId"],
                        codec="vp8",
                    )

    async def _handle_add_video_stream(self, response: dict[str, Any]) -> None:
        """Handle add video stream notification and auto-subscribe.

        The message type 'on_add_video_stream' itself implies this is a video
        stream — do not require an additional 'video' boolean field.
        If the user is already online, subscribe immediately.
        """
        message = response.get("_message", {})
        uid = message.get("uid")
        ssrc_id = message.get("ssrcId")
        rtx_ssrc_id = message.get("rtxSsrcId")
        cname = message.get("cname")

        if uid:
            _LOGGER.info(
                "Video stream added - uid: %s, ssrcId: %s, rtxSsrcId: %s, cname: %s",
                uid,
                ssrc_id,
                rtx_ssrc_id,
                cname,
            )

            # Store stream info
            self._video_streams[uid] = {
                "ssrcId": ssrc_id,
                "rtxSsrcId": rtx_ssrc_id,
                "cname": cname,
                "subscribed": False,
            }

            # Auto-subscribe if user is already online
            if uid in self._online_users and self._websocket:
                _LOGGER.info("User %s already online, subscribing to video stream now", uid)
                self._video_streams[uid]["subscribed"] = True
                await self._send_subscribe(stream_id=uid, ssrc_id=ssrc_id, codec="vp8")

    async def _handle_user_offline(self, response: dict[str, Any]) -> None:
        """Handle user offline notification.

        Send unsubscribe for the user's stream and clean up tracking state.
        """
        message = response.get("_message", {})
        uid = message.get("uid")
        reason = message.get("reason", "unknown")
        if uid:
            _LOGGER.info("User %s went offline (reason: %s)", uid, reason)
            self._online_users.discard(uid)

            # Unsubscribe if we had an active subscription
            if uid in self._video_streams:
                if self._video_streams[uid].get("subscribed") and self._websocket:
                    await self._send_unsubscribe(stream_id=uid)
                del self._video_streams[uid]

    async def _send_unsubscribe(
        self,
        stream_id: int,
        p2p_id: int = 1,
    ) -> None:
        """Send unsubscribe message to Agora.

        Args:
            stream_id: Stream ID (the uid of the remote user)
            p2p_id: P2P ID (default: 1)

        """
        if not self._websocket:
            _LOGGER.error("Cannot send unsubscribe: websocket not connected")
            return

        message_id = secrets.token_hex(3)
        message = {
            "_id": message_id,
            "_type": "unsubscribe",
            "_message": {
                "p2p_id": p2p_id,
                "ortc": [],
                "stream_id": stream_id,
            },
        }

        _LOGGER.info("Sending unsubscribe message: %s", message)
        await self._websocket.send(json.dumps(message))

    def _create_join_message(
        self,
        agora_data: StreamSubscriptionResponse,
        offer_sdp: str,
        stored_sdp_info: SdpInfo,
        ortc_info: dict[str, Any],
        agora_response: AgoraResponse,
        session_id: str,
    ) -> dict[str, Any]:
        """Create join_v3 message for Agora WebSocket."""
        message_id = secrets.token_hex(3)  # 6 characters
        process_id = f"process-{secrets.token_hex(4)}-{secrets.token_hex(2)}-{secrets.token_hex(2)}-{secrets.token_hex(2)}-{secrets.token_hex(6)}"

        return {
            "_id": message_id,
            "_type": "join_v3",
            "_message": {
                "p2p_id": 1,
                "session_id": session_id,
                "app_id": agora_data.appid,
                "channel_key": agora_data.token,
                "channel_name": agora_data.channelName,
                "sdk_version": "4.24.0",
                "browser": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
                "process_id": process_id,
                "mode": "live",
                "codec": "vp8",
                "role": "host",
                "has_changed_gateway": False,
                "ap_response": agora_response.to_ap_response(4096),
                "extend": "",
                "details": {},
                "features": {"rejoin": True},
                "attributes": {
                    "userAttributes": {
                        "enableAudioMetadata": False,
                        "enableAudioPts": False,
                        "enablePublishedUserList": True,
                        "maxSubscription": 50,
                        "enableUserLicenseCheck": True,
                        "enableRTX": True,
                        "enableInstantVideo": False,
                        "enableDataStream2": False,
                        "enableAutFeedback": True,
                        "enableUserAutoRebalanceCheck": True,
                        "enableXR": True,
                        "enableLossbasedBwe": True,
                        "enableAutCC": True,
                        "enablePreallocPC": False,
                        "enablePubTWCC": False,
                        "enableSubTWCC": True,
                        "enablePubRTX": True,
                        "enableSubRTX": True,
                    }
                },
                "join_ts": int(time.time() * 1000),
                "ortc": ortc_info,
            },
        }

    async def _send_set_client_role(self, role: str = "audience", level: int = 1) -> None:
        """Send set_client_role message to Agora.

        Args:
            role: Client role - "audience" or "host"
            level: Role level (default: 1)

        """
        if not self._websocket:
            _LOGGER.error("Cannot send set_client_role: websocket not connected")
            return

        message_id = secrets.token_hex(3)
        message = {
            "_id": message_id,
            "_type": "set_client_role",
            "_message": {
                "role": role,
                "level": level,
                "client_ts": int(time.time() * 1000),
            },
        }

        _LOGGER.info("Sending set_client_role message: %s", message)
        await self._websocket.send(json.dumps(message))

    async def _send_subscribe(
        self,
        stream_id: int,
        ssrc_id: int,
        codec: str = "vp8",
        stream_type: str = "video",
        mode: str = "live",
        p2p_id: int = 1,
        twcc: bool = True,
        rtx: bool = True,
        extend: str = "",
    ) -> None:
        """Send subscribe message to Agora.

        Args:
            stream_id: Stream ID (usually the uid)
            ssrc_id: SSRC ID from on_add_video_stream
            codec: Video codec (default: "vp8")
            stream_type: Stream type (default: "video")
            mode: Mode (default: "live")
            p2p_id: P2P ID (default: 1)
            twcc: Enable transport-wide congestion control
            rtx: Enable retransmission
            extend: Extended info

        """
        if not self._websocket:
            _LOGGER.error("Cannot send subscribe: websocket not connected")
            return

        message_id = secrets.token_hex(3)
        message = {
            "_id": message_id,
            "_type": "subscribe",
            "_message": {
                "stream_id": stream_id,
                "stream_type": stream_type,
                "mode": mode,
                "codec": codec,
                "p2p_id": p2p_id,
                "twcc": twcc,
                "rtx": rtx,
                "extend": extend,
                "ssrcId": ssrc_id,
            },
        }

        _LOGGER.info("Sending subscribe message: %s", message)
        await self._websocket.send(json.dumps(message))

    def _convert_candidates_to_ortc(self) -> list[dict[str, Any]]:
        """Convert collected ICE candidates to ORTC format for join message.

        Returns:
            List of candidate dictionaries in ORTC format

        """
        ortc_candidates = []

        for candidate in self.candidates:
            cand_str = candidate.candidate
            if not cand_str:
                continue

            # Remove \"candidate:\" prefix if present
            if cand_str.startswith("candidate:"):
                cand_str = cand_str[10:]

            parts = cand_str.split()
            if len(parts) < 8:
                _LOGGER.warning("Invalid candidate format: %s", candidate.candidate)
                continue

            try:
                foundation = parts[0]
                protocol = parts[2]
                priority = int(parts[3])
                ip = parts[4]
                port = int(parts[5])
                cand_type = parts[7]

                # Build candidate object in ORTC format
                ortc_candidates.append(
                    {
                        "foundation": foundation,
                        "ip": ip,
                        "port": port,
                        "priority": priority,
                        "protocol": protocol,
                        "type": cand_type,
                    }
                )

            except (ValueError, IndexError) as ex:
                _LOGGER.error("Failed to parse candidate %s: %s", candidate.candidate, ex)
                continue

        _LOGGER.info("Converted %d candidates to ORTC format", len(ortc_candidates))
        return ortc_candidates

    @staticmethod
    def _add_candidates_to_sdp(sdp: str, candidates: list[RTCIceCandidateInit]) -> str:
        """Add ICE candidates to SDP.

        Args:
            sdp: Original SDP string
            candidates: List of ICE candidates (dict or string format)

        Returns:
            Modified SDP with candidates added

        """
        sdp_lines = sdp.split("\n")
        result_lines = []
        in_media_section = False

        for i, line in enumerate(sdp_lines):
            # Check if we're entering a media section
            if line.startswith("m="):
                in_media_section = True

            # Check if we're at the end of a media section (next m= line or end of SDP)
            next_is_media = i + 1 < len(sdp_lines) and sdp_lines[i + 1].startswith("m=")
            is_last_line = i == len(sdp_lines) - 1

            result_lines.append(line)

            # Add candidates at the end of each media section
            if in_media_section and (next_is_media or is_last_line):
                for cand in candidates:
                    cand_str = cand.candidate

                    if not cand_str:
                        continue

                    # Normalize to proper format (ensure it starts with "a=candidate:")
                    if not cand_str.startswith("a="):
                        if cand_str.startswith("candidate:"):
                            cand_str = "a=" + cand_str
                        else:
                            cand_str = "a=candidate:" + cand_str

                    result_lines.append(cand_str)

        return "\n".join(result_lines)

    @staticmethod
    def _parse_offer_sdp(offer_sdp: str) -> SdpInfo | None:
        """Parse offer SDP to extract capabilities and parameters using sdp_transform."""
        try:
            # Parse SDP using sdp_transform
            parsed_sdp = sdp_parse(offer_sdp)
            _LOGGER.info("Parsed SDP structure: %s", parsed_sdp)

            # Extract fingerprint
            fingerprint = ""
            if "fingerprint" in parsed_sdp:
                fingerprint = parsed_sdp["fingerprint"]["hash"]
            else:
                # Check in media sections
                for media in parsed_sdp.get("media", []):
                    if "fingerprint" in media:
                        fingerprint = media["fingerprint"]["hash"]
                        break

            # Extract ICE parameters
            ice_ufrag = parsed_sdp.get("iceUfrag", "")
            ice_pwd = parsed_sdp.get("icePwd", "")

            # Check in media sections if not found at top level
            if not ice_ufrag or not ice_pwd:
                for media in parsed_sdp.get("media", []):
                    if not ice_ufrag and "iceUfrag" in media:
                        ice_ufrag = media["iceUfrag"]
                    if not ice_pwd and "icePwd" in media:
                        ice_pwd = media["icePwd"]
                    if ice_ufrag and ice_pwd:
                        break

            audio_codecs = []
            video_codecs = []
            audio_extensions = []
            video_extensions = []

            audio_direction = "sendrecv"
            video_direction = "sendrecv"

            # Process media sections
            for media in parsed_sdp.get("media", []):
                media_type = media.get("type")

                # capture direction per media so answer generator can choose complementary dir
                dir_val = media.get("direction", "sendrecv")

                if media_type == "audio":
                    audio_direction = dir_val
                elif media_type == "video":
                    video_direction = dir_val

                # Process RTP codecs
                for rtp in media.get("rtp", []):
                    codec_entry = {
                        "payloadType": rtp["payload"],
                        "rtpMap": {
                            "encodingName": rtp["codec"],
                            "clockRate": rtp["rate"],
                        },
                        "rtcpFeedbacks": [],
                    }

                    # Add encoding parameters if present
                    if "encoding" in rtp:
                        codec_entry["rtpMap"]["encodingParameters"] = rtp["encoding"]

                    # Find fmtp parameters for this payload type
                    fmtp_params = {}
                    for fmtp in media.get("fmtp", []):
                        if fmtp["payload"] == rtp["payload"]:
                            # Parse config string into parameters
                            config = fmtp.get("config", "")
                            if config:
                                params = {}
                                for param_pair in config.split(";"):
                                    if "=" in param_pair:
                                        key, value = param_pair.split("=", 1)
                                        params[key.strip()] = value.strip()
                                    else:
                                        # Handle cases like "111/111" for RED codec
                                        params[param_pair.strip()] = None
                                if params:
                                    fmtp_params = params
                            break

                    # Add fmtp if found
                    if fmtp_params:
                        codec_entry["fmtp"] = {"parameters": fmtp_params}

                    # Process RTCP feedback from SDP
                    rtcp_feedbacks = []
                    for rtcp_fb in media.get("rtcpFb", []):
                        if rtcp_fb.get("payload") == rtp["payload"]:
                            feedback = {"type": rtcp_fb["type"]}
                            if "subtype" in rtcp_fb:
                                feedback["parameter"] = rtcp_fb["subtype"]
                            rtcp_feedbacks.append(feedback)

                    # Add default RTCP feedback based on media type and codec if none found
                    if not rtcp_feedbacks:
                        codec_name = rtp["codec"].upper()
                        if media_type == "video":
                            if codec_name in ["VP8", "VP9", "H264", "AV1"]:
                                rtcp_feedbacks = [
                                    {"type": "goog-remb"},
                                    {"type": "transport-cc"},
                                    {"type": "ccm", "parameter": "fir"},
                                    {"type": "nack"},
                                    {"type": "nack", "parameter": "pli"},
                                    {"type": "rrtr"},
                                ]
                            elif codec_name == "RTX":
                                rtcp_feedbacks = [{"type": "rrtr"}]
                            else:
                                rtcp_feedbacks = [{"type": "rrtr"}]
                        elif media_type == "audio":
                            rtcp_feedbacks = [{"type": "rrtr"}]
                            if codec_name == "OPUS":
                                rtcp_feedbacks.append({"type": "transport-cc"})

                    codec_entry["rtcpFeedbacks"] = rtcp_feedbacks

                    if media_type == "video":
                        video_codecs.append(codec_entry)
                    elif media_type == "audio":
                        audio_codecs.append(codec_entry)

                # Process extensions
                for ext in media.get("ext", []):
                    ext_entry = {
                        "entry": ext["value"],
                        "extensionName": ext["uri"],
                    }

                    # Map common extension names to match Agora's format
                    uri_mappings = {
                        "urn:ietf:params:rtp-hdrext:ssrc-audio-level": "urn:ietf:params:rtp-hdrext:ssrc-audio-level",
                        "http://www.webrtc.org/experiments/rtp-hdrext/abs-send-time": "http://www.webrtc.org/experiments/rtp-hdrext/abs-send-time",
                        "http://www.ietf.org/id/draft-holmer-rmcat-transport-wide-cc-extensions-01": "http://www.ietf.org/id/draft-holmer-rmcat-transport-wide-cc-extensions-01",
                        "urn:ietf:params:rtp-hdrext:sdes:mid": "urn:ietf:params:rtp-hdrext:sdes:mid",
                        "urn:ietf:params:rtp-hdrext:toffset": "urn:ietf:params:rtp-hdrext:toffset",
                        "urn:3gpp:video-orientation": "urn:3gpp:video-orientation",
                        "http://www.webrtc.org/experiments/rtp-hdrext/playout-delay": "http://www.webrtc.org/experiments/rtp-hdrext/playout-delay",
                        "http://www.webrtc.org/experiments/rtp-hdrext/video-content-type": "http://www.webrtc.org/experiments/rtp-hdrext/video-content-type",
                        "http://www.webrtc.org/experiments/rtp-hdrext/video-timing": "http://www.webrtc.org/experiments/rtp-hdrext/video-timing",
                        "http://www.webrtc.org/experiments/rtp-hdrext/color-space": "http://www.webrtc.org/experiments/rtp-hdrext/color-space",
                        "urn:ietf:params:rtp-hdrext:sdes:rtp-stream-id": "urn:ietf:params:rtp-hdrext:sdes:rtp-stream-id",
                        "urn:ietf:params:rtp-hdrext:sdes:repaired-rtp-stream-id": "urn:ietf:params:rtp-hdrext:sdes:repaired-rtp-stream-id",
                    }

                    # Use mapped URI if available
                    ext_entry["extensionName"] = uri_mappings.get(ext["uri"], ext["uri"])

                    if media_type == "audio":
                        audio_extensions.append(ext_entry)
                    elif media_type == "video":
                        video_extensions.append(ext_entry)

            # Extract ICE candidates from SDP
            ice_candidates = []
            for media in parsed_sdp.get("media", []):
                for candidate in media.get("candidates", []):
                    # Convert sdp_transform candidate format to ORTC format
                    ice_candidate = {
                        "foundation": candidate.get("foundation", ""),
                        "protocol": candidate.get("transport", "udp"),
                        "priority": candidate.get("priority", 0),
                        "ip": candidate.get("ip", ""),
                        "port": candidate.get("port", 0),
                        "type": candidate.get("type", "host"),
                    }
                    # Add optional fields if present
                    if "generation" in candidate:
                        ice_candidate["generation"] = candidate["generation"]
                    if "raddr" in candidate:
                        ice_candidate["relatedAddress"] = candidate["raddr"]
                    if "rport" in candidate:
                        ice_candidate["relatedPort"] = candidate["rport"]
                    if "tcptype" in candidate:
                        ice_candidate["tcpType"] = candidate["tcptype"]

                    ice_candidates.append(ice_candidate)

            # Extract extmap-allow-mixed from session level
            extmap_allow_mixed = parsed_sdp.get("extmapAllowMixed", False)

            # Extract setup role from first media section (usually same for all)
            setup_role = "actpass"  # default
            for media in parsed_sdp.get("media", []):
                if "setup" in media:
                    setup_role = media["setup"]
                    break

            return SdpInfo(
                parsed_sdp,
                fingerprint=fingerprint,
                ice_ufrag=ice_ufrag,
                ice_pwd=ice_pwd,
                audio_codecs=audio_codecs,
                video_codecs=video_codecs,
                audio_extensions=audio_extensions,
                video_extensions=video_extensions,
                audio_direction=audio_direction,
                video_direction=video_direction,
                ice_candidates=ice_candidates,
                extmap_allow_mixed=extmap_allow_mixed,
                setup_role=setup_role,
            )

        except (ValueError, IndexError, KeyError) as ex:
            _LOGGER.error("Failed to parse offer SDP with sdp_transform: %s", ex)
            return None

    def _generate_answer_sdp(self, ortc: dict[str, Any], sdp_info: SdpInfo) -> str | None:
        """Generate SDP answer from ORTC parameters."""
        try:
            from collections import defaultdict
            import secrets

            ice_params = ortc.get("iceParameters", {})
            dtls_params = ortc.get("dtlsParameters", {})
            # Server may return caps under 'sendrecv', 'recv', or 'send' keys
            rtp_capabilities = ortc.get("rtpCapabilities", {})
            rtp_caps = (
                rtp_capabilities.get("sendrecv")
                or rtp_capabilities.get("recv")
                or rtp_capabilities.get("send")
                or rtp_capabilities
            )

            _LOGGER.debug("ICE params: %s", ice_params)
            _LOGGER.debug("DTLS params: %s", dtls_params)
            _LOGGER.debug("RTP caps: %s", rtp_caps)

            # Extract ICE credentials from ORTC (these are OUR credentials for the answer)
            ortc_candidates = ice_params.get("candidates", []) or []
            ice_ufrag = ice_params.get("iceUfrag", "") or ""
            ice_pwd = ice_params.get("icePwd", "") or ""

            _LOGGER.info(
                "Answer SDP will use ICE ufrag: %s, " "and will include %d candidates from Agora response",
                ice_ufrag,
                len(ortc_candidates),
            )

            # fallback credentials
            if not ice_ufrag:
                ice_ufrag = secrets.token_hex(4)
                _LOGGER.warning("Using fallback ICE ufrag: %s", ice_ufrag)
            if not ice_pwd:
                ice_pwd = secrets.token_hex(16)
                _LOGGER.warning("Using fallback ICE pwd")

            # Extract DTLS fingerprint
            fingerprints = dtls_params.get("fingerprints", []) or []
            fingerprint = ""
            if fingerprints:
                fp = fingerprints[0]
                fingerprint = f"{fp.get('algorithm', 'sha-256')} {fp.get('fingerprint', '')}"
            if not fingerprint:
                fallback_fingerprint = ":".join([secrets.token_hex(1).upper() for _ in range(32)])
                fingerprint = f"sha-256 {fallback_fingerprint}"
                _LOGGER.warning("Using fallback fingerprint")

            # Build candidates from ORTC response for initial connectivity
            candidates_by_mid = defaultdict(list)
            for i, c in enumerate(ortc_candidates):
                foundation = c.get("foundation", f"candidate{i}")
                protocol = c.get("protocol", "udp")
                priority = f"{c.get('priority', 2103266323)}"
                ip = c.get("ip", "")
                port = f"{c.get('port', 0)}"
                ctype = c.get("type", "host")
                cand_line = f"a=candidate:{foundation} 1 {protocol} {priority} {ip} {port} typ {ctype}"
                if c.get("generation") is not None:
                    cand_line += f" generation {c.get('generation')}"
                candidates_by_mid["*"].append(cand_line)
                _LOGGER.debug("Built candidate line: %s", cand_line)

            # Extract codec lists and extensions from ORTC
            video_codecs = rtp_caps.get("videoCodecs", []) or []
            audio_codecs = rtp_caps.get("audioCodecs", []) or []
            video_extensions = rtp_caps.get("videoExtensions", []) or []
            audio_extensions = rtp_caps.get("audioExtensions", []) or []

            def answer_direction_for_offer(offer_dir: str) -> str:
                if offer_dir == "sendonly":
                    return "recvonly"
                if offer_dir == "recvonly":
                    return "sendonly"
                if offer_dir == "sendrecv":
                    return "sendrecv"
                return "inactive"

            # Parse media from offer - MUST preserve exact order for answer
            media = sdp_info.parsed_sdp.get("media", []) or []

            # Build BUNDLE list from offer
            bundle_group = sdp_info.parsed_sdp.get("groups", [{}])[0] if sdp_info.parsed_sdp.get("groups") else {}
            bundle_mids = bundle_group.get("mids", "0 1") if bundle_group else "0 1"

            # Determine answer setup role based on offer
            # Working SDK Answer shows setup:active
            answer_setup = "active"

            # build base sdp header
            # f"o=- {sdp_info.parsed_sdp['origin']['sessionId']} {sdp_info.parsed_sdp['origin']['sessionVersion']} IN IP4 127.0.0.1",
            sdp_lines = [
                "v=0",
                "o=- 0 0 IN IP4 127.0.0.1",
                "s=AgoraGateway",
                "t=0 0",
                f"a=group:BUNDLE {bundle_mids}",
            ]

            # Add ice-lite to session level as seen in working SDK
            sdp_lines.append("a=ice-lite")

            # Add extmap-allow-mixed if present in answer
            if sdp_info.extmap_allow_mixed:
                sdp_lines.append("a=extmap-allow-mixed")

            sdp_lines.append("a=msid-semantic: WMS")

            # Generate media sections in EXACT SAME ORDER as offer
            # The answer MUST match the offer's m-line order
            for idx, m in enumerate(media):
                mtype = m.get("type", "audio")
                offer_dir = m.get("direction", "sendonly")
                answer_dir = answer_direction_for_offer(offer_dir)
                mid = str(m.get("mid", str(idx)))

                # Select codecs for this media type
                if mtype == "audio":
                    codecs = audio_codecs
                    extensions = audio_extensions
                else:
                    codecs = video_codecs
                    extensions = video_extensions

                # Build payload type list
                payload_types = [str(codec.get("payloadType")) for codec in codecs]
                payloads_str = " ".join(payload_types)

                # media header
                sdp_lines.append(f"m={mtype} 9 UDP/TLS/RTP/SAVPF {payloads_str}")
                sdp_lines.append("c=IN IP4 127.0.0.1")
                sdp_lines.append("a=rtcp:9 IN IP4 0.0.0.0")
                sdp_lines.append(f"a=ice-ufrag:{ice_ufrag}")
                sdp_lines.append(f"a=ice-pwd:{ice_pwd}")
                sdp_lines.append("a=ice-options:trickle")
                sdp_lines.append(f"a=fingerprint:{fingerprint}")
                sdp_lines.append(f"a=setup:{answer_setup}")
                sdp_lines.append(f"a=mid:{mid}")

                # Add candidates from Agora response
                for cl in candidates_by_mid.get("*", []):
                    sdp_lines.append(cl)

                # Add RTP extensions - MUST use offer's extension IDs
                # Build mapping from offer's extension URIs to their IDs
                offer_ext_map = {}
                if mtype == "audio":
                    for ext in sdp_info.audio_extensions:
                        offer_ext_map[ext.get("extensionName")] = ext.get("entry")
                else:
                    for ext in sdp_info.video_extensions:
                        offer_ext_map[ext.get("extensionName")] = ext.get("entry")

                # Add extensions using offer's IDs for matching URIs
                for ext in extensions:
                    ext_name = ext.get("extensionName")
                    if not ext_name:
                        continue

                    # Use the offer's extension ID if this extension was in the offer
                    if ext_name in offer_ext_map:
                        entry = offer_ext_map[ext_name]
                        sdp_lines.append(f"a=extmap:{entry} {ext_name}")
                    # Otherwise skip this extension (not negotiated in offer)

                sdp_lines.append(f"a={answer_dir}")
                sdp_lines.append("a=rtcp-mux")
                sdp_lines.append("a=rtcp-rsize")

                # Add codec details (rtpmap, rtcp-fb, fmtp)
                for codec in codecs:
                    pt = codec.get("payloadType")
                    rtp_map = codec.get("rtpMap", {})
                    encoding_name = rtp_map.get("encodingName", "")
                    clock_rate = rtp_map.get("clockRate", 90000)
                    encoding_params = rtp_map.get("encodingParameters")

                    # rtpmap line
                    if encoding_params:
                        sdp_lines.append(f"a=rtpmap:{pt} {encoding_name}/{clock_rate}/{encoding_params}")
                    else:
                        sdp_lines.append(f"a=rtpmap:{pt} {encoding_name}/{clock_rate}")

                    # rtcp-fb lines
                    for fb in codec.get("rtcpFeedbacks", []):
                        fb_type = fb.get("type")
                        fb_param = fb.get("parameter")
                        if fb_param:
                            sdp_lines.append(f"a=rtcp-fb:{pt} {fb_type} {fb_param}")
                        else:
                            sdp_lines.append(f"a=rtcp-fb:{pt} {fb_type}")

                    # fmtp line
                    fmtp = codec.get("fmtp", {})
                    if fmtp:
                        params = fmtp.get("parameters", {})
                        if params:
                            param_str = ";".join([f"{k}={v}" for k, v in params.items()])
                            sdp_lines.append(f"a=fmtp:{pt} {param_str}")

                # Working SDK answer DOES NOT include a=ssrc for audience/receiver section
                # Omit SSRC for receiver role

                # Append candidates from Agora response for trickle ICE initialization
                # These are the TURN/STUN candidates provided by Agora in the join_success response
                specific = candidates_by_mid.get(mid, []) + candidates_by_mid.get(str(idx), [])
                for cl in specific:
                    sdp_lines.append(cl)

                if specific:
                    _LOGGER.info(
                        "Added %d ICE candidates to media section %s (type=%s)",
                        len(specific),
                        mid,
                        mtype,
                    )

            generated_sdp = "\r\n".join(sdp_lines) + "\r\n"
            # _LOGGER.info("Generated SDP lines count: %s", len(sdp_lines))
            # _LOGGER.debug("Generated SDP content: %s", generated_sdp)

            if self._validate_sdp(generated_sdp):
                return generated_sdp
            _LOGGER.error("Generated SDP failed validation")
            return None

        except (KeyError, ValueError, AttributeError) as ex:
            _LOGGER.error("Failed to generate answer SDP: %s", ex)
            return None

    def _validate_sdp(self, sdp: str) -> bool:
        """Validate SDP format to ensure it's parseable by WebRTC."""
        if not sdp or len(sdp.strip()) == 0:
            _LOGGER.error("SDP is empty")
            return False

        lines = sdp.split("\r\n")
        has_version = False
        has_origin = False
        has_session_name = False
        has_timing = False
        m_line_count = 0

        for line in lines:
            line = line.strip()
            if not line:
                continue

            if line.startswith("v="):
                has_version = True
            elif line.startswith("o="):
                has_origin = True
            elif line.startswith("s="):
                has_session_name = True
            elif line.startswith("t="):
                has_timing = True
            elif line.startswith("m="):
                m_line_count += 1

        if not has_version:
            _LOGGER.error("SDP missing version line (v=)")
            return False
        if not has_origin:
            _LOGGER.error("SDP missing origin line (o=)")
            return False
        if not has_session_name:
            _LOGGER.error("SDP missing session name line (s=)")
            return False
        if not has_timing:
            _LOGGER.error("SDP missing timing line (t=)")
            return False
        if m_line_count < 2:
            _LOGGER.error("SDP has %s m-lines, expected 2 (audio + video)", m_line_count)
            return False

        _LOGGER.debug("SDP validation passed: %s m-lines found", m_line_count)
        return True

    def _generate_fallback_sdp(self) -> str:
        """Generate a basic fallback SDP answer."""
        _LOGGER.info("Generating fallback SDP with default parameters")

        # Generate basic parameters
        ice_ufrag = secrets.token_hex(4)
        ice_pwd = secrets.token_hex(16)
        fallback_fingerprint = ":".join([secrets.token_hex(1).upper() for _ in range(32)])
        fingerprint = f"sha-256 {fallback_fingerprint}"

        # Default codec payload types
        opus_payload = 109
        vp8_payload = 120

        # Build fallback SDP answer
        sdp_lines = [
            "v=0",
            f"o=- {secrets.randbelow(2**63)} 2 IN IP4 127.0.0.1",
            "s=-",
            "t=0 0",
            "a=group:BUNDLE 0 1",
            "a=msid-semantic: WMS",
            "",
            # Audio m-line
            f"m=audio 9 UDP/TLS/RTP/SAVPF {opus_payload}",
            "c=IN IP4 0.0.0.0",
            "a=rtcp:9 IN IP4 0.0.0.0",
            f"a=ice-ufrag:{ice_ufrag}",
            f"a=ice-pwd:{ice_pwd}",
            "a=ice-options:trickle",
            f"a=fingerprint:{fingerprint}",
            "a=setup:active",
            "a=mid:0",
            "a=sendrecv",
            "a=rtcp-mux",
            f"a=rtpmap:{opus_payload} opus/48000/2",
            "",
            # Video m-line
            f"m=video 9 UDP/TLS/RTP/SAVPF {vp8_payload}",
            "c=IN IP4 0.0.0.0",
            "a=rtcp:9 IN IP4 0.0.0.0",
            f"a=ice-ufrag:{ice_ufrag}",
            f"a=ice-pwd:{ice_pwd}",
            "a=ice-options:trickle",
            f"a=fingerprint:{fingerprint}",
            "a=setup:active",
            "a=mid:1",
            "a=sendrecv",
            "a=rtcp-mux",
            f"a=rtpmap:{vp8_payload} VP8/90000",
        ]

        generated_sdp = "\r\n".join(sdp_lines) + "\r\n"
        # _LOGGER.debug("Generated fallback SDP: %s", generated_sdp)

        # Validate fallback SDP
        if self._validate_sdp(generated_sdp):
            return generated_sdp
        _LOGGER.error("Fallback SDP failed validation")
        # Return a minimal valid SDP as last resort
        return self._generate_minimal_sdp()

    def _generate_minimal_sdp(self) -> str:
        """Generate minimal valid SDP as last resort."""
        _LOGGER.warning("Generating minimal SDP as last resort")

        ice_ufrag = secrets.token_hex(4)
        ice_pwd = secrets.token_hex(16)

        return (
            "v=0\r\n"
            f"o=- {secrets.randbelow(2**63)} 2 IN IP4 127.0.0.1\r\n"
            "s=-\r\n"
            "t=0 0\r\n"
            "a=group:BUNDLE 0 1\r\n"
            "a=msid-semantic: WMS\r\n"
            "m=audio 9 UDP/TLS/RTP/SAVPF 109\r\n"
            "c=IN IP4 0.0.0.0\r\n"
            f"a=ice-ufrag:{ice_ufrag}\r\n"
            f"a=ice-pwd:{ice_pwd}\r\n"
            "a=fingerprint:sha-256 00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00\r\n"
            "a=setup:active\r\n"
            "a=mid:0\r\n"
            "a=sendrecv\r\n"
            "a=rtcp-mux\r\n"
            "a=rtpmap:109 opus/48000/2\r\n"
            "m=video 9 UDP/TLS/RTP/SAVPF 120\r\n"
            "c=IN IP4 0.0.0.0\r\n"
            f"a=ice-ufrag:{ice_ufrag}\r\n"
            f"a=ice-pwd:{ice_pwd}\r\n"
            "a=fingerprint:sha-256 00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00\r\n"
            "a=setup:active\r\n"
            "a=mid:1\r\n"
            "a=sendrecv\r\n"
            "a=rtcp-mux\r\n"
            "a=rtpmap:120 VP8/90000\r\n"
        )

    async def _get_agora_edge_services(self, agora_data: StreamSubscriptionResponse) -> ResponseInfo | None:
        """Get Agora edge services information."""
        app_id = agora_data.appid
        channel_name = agora_data.channelName
        token = agora_data.token
        uid = int(agora_data.uid)

        # Generate required IDs for the API call
        client_ts = int(time.time() * 1000)
        opid = secrets.randbelow(2**31)
        sid = secrets.token_hex(16).upper()

        # Create the request payload
        request_payload = {
            "appid": app_id,
            "client_ts": client_ts,
            "opid": opid,
            "sid": sid,
            "request_bodies": [
                {
                    "uri": 22,
                    "buffer": {
                        "cname": channel_name,
                        "detail": {"11": "CN,GLOBAL", "17": "1", "22": "CN,GLOBAL"},
                        "key": token,
                        "service_ids": [11, 26],
                        "uid": uid,
                    },
                }
            ],
        }

        # Create multipart form data using aiohttp.MultipartWriter
        writer = aiohttp.MultipartWriter("form-data")
        part = writer.append(json.dumps(request_payload))
        part.set_content_disposition("form-data", name="request")

        headers = {
            "User-Agent": "Home Assistant WebRTC",
        }

        api_url = "https://webrtc2-ap-web-1.agora.io/api/v2/transpond/webrtc?v=2"

        try:
            async with (
                aiohttp.ClientSession() as session,
                session.post(
                    api_url,
                    data=writer,
                    headers=headers,
                    timeout=aiohttp.ClientTimeout(total=10),
                ) as response,
            ):
                if response.status != 200:
                    _LOGGER.error("Agora API returned status %s", response.status)
                    raise aiohttp.ClientError(f"API returned status {response.status}")

                # Read response as JSON
                response_text = await response.text()
                _LOGGER.debug("Agora API raw response: %s", response_text)

                response_data = json.loads(response_text)
                _LOGGER.debug("Agora API parsed response: %s", response_data)

                # Extract edge services from response
                response_bodies = response_data.get("response_body", [])
                for body in reversed(response_bodies):
                    buffer = body.get("buffer", {})
                    if buffer and buffer.get("flag") == 4096:
                        edges_services = buffer.get("edges_services", [])
                        if edges_services:
                            es = next(iter(edges_services), None)
                            return ResponseInfo(
                                code=buffer["code"],
                                addresses=[
                                    AddressEntry(
                                        ip=es["ip"],
                                        port=es["port"],
                                        ticket=buffer["cert"],
                                    )
                                ],
                                server_ts=response_data["enter_ts"],
                                uid=buffer["uid"],
                                cid=buffer["cid"],
                                cname=buffer["cname"],
                                detail={
                                    **buffer.get("detail", {}),
                                    **response_data.get("detail", {}),
                                },
                                flag=buffer["flag"],
                                opid=response_data["opid"],
                                cert=buffer["cert"],
                            )

                # Fallback if no edge services found
                _LOGGER.warning("No edge services found in Agora API response, using fallback")
                raise aiohttp.ClientError("No edge services available")

        except (aiohttp.ClientError, json.JSONDecodeError) as ex:
            _LOGGER.error("Failed to get Agora edge services: %s", ex)
            return None

    @property
    def is_connected(self) -> bool:
        """Return whether WebSocket is connected."""
        return self._connection_state == "CONNECTED"

    async def disconnect(self) -> None:
        """Disconnect from WebSocket and clean up background tasks."""
        # Cancel background tasks
        if self._ping_task and not self._ping_task.done():
            self._ping_task.cancel()
            self._ping_task = None
        if self._message_loop_task and not self._message_loop_task.done():
            self._message_loop_task.cancel()
            self._message_loop_task = None

        # Close WebSocket
        if self._websocket:
            try:
                await self._websocket.close()
            except Exception:  # noqa: BLE001
                pass
            self._websocket = None

        # Clear token state
        self._rejoin_token = None
        self._connection_state = "DISCONNECTED"

    def add_ice_candidate(self, candidate: RTCIceCandidateInit) -> None:
        self.candidates.append(candidate)

    @staticmethod
    def is_ipv4(ip_string) -> bool:
        """Checks if a given string is a valid IPv4 address.

        Args:
            ip_string (str): The string to validate.

        Returns:
            bool: True if the string is a valid IPv4 address, False otherwise.

        """
        try:
            # Attempt to create an IPv4Address object
            ipaddress.IPv4Address(ip_string)
            return True
        except ipaddress.AddressValueError:
            # If it's not a valid IPv4 address, an exception will be raised
            return False
        except ValueError:
            # Catch other potential ValueErrors (e.g., if it's an IPv6 address)
            # and ensure it's specifically an IPv4Address error.
            # This is a more robust way to handle potential edge cases.
            return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymammotion/proto/mctrl_nav.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pymammotion.proto import common_pb2 as pymammotion_dot_proto_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!pymammotion/proto/mctrl_nav.proto\x1a\x1epymammotion/proto/common.proto\"\'\n\x0bNavLatLonUp\x12\x0b\n\x03lat\x18\x01 \x01(\x01\x12\x0b\n\x03lon\x18\x02 \x01(\x01\"!\n\x0eNavBorderState\x12\x0f\n\x07\x62\x64state\x18\x01 \x01(\x05\"\xc9\x01\n\x08NavPosUp\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\x0e\n\x06status\x18\x03 \x01(\x05\x12\x0e\n\x06toward\x18\x04 \x01(\x05\x12\r\n\x05stars\x18\x05 \x01(\x05\x12\x0b\n\x03\x61ge\x18\x06 \x01(\x02\x12\x11\n\tlatStddev\x18\x07 \x01(\x02\x12\x11\n\tlonStddev\x18\x08 \x01(\x02\x12\x11\n\tl2dfStars\x18\t \x01(\x05\x12\x0f\n\x07posType\x18\n \x01(\x05\x12\x0f\n\x07\x63HashId\x18\x0b \x01(\x03\x12\x10\n\x08posLevel\x18\x0c \x01(\x05\":\n\x13NavBorderDataGetAck\x12\r\n\x05jobId\x18\x01 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x02 \x01(\x05\"Z\n\x15NavObstiBorderDataGet\x12\x15\n\robstacleIndex\x18\x01 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x02 \x01(\x05\x12\x14\n\x0cobstaclesLen\x18\x03 \x01(\x05\"G\n\x18NavObstiBorderDataGetAck\x12\x15\n\robstacleIndex\x18\x01 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x02 \x01(\x05\"d\n\x0eNavCHlLineData\x12\x12\n\nstartJobRI\x18\x01 \x01(\x05\x12\x10\n\x08\x65ndJobRI\x18\x02 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x03 \x01(\x05\x12\x16\n\x0e\x63hannelLineLen\x18\x04 \x01(\x05\"O\n\x11NavCHlLineDataAck\x12\x12\n\nstartJobRI\x18\x01 \x01(\x05\x12\x10\n\x08\x65ndJobRI\x18\x02 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x03 \x01(\x05\"\x7f\n\x0bNavTaskInfo\x12\x0c\n\x04\x61rea\x18\x01 \x01(\x05\x12\x0c\n\x04time\x18\x02 \x01(\x05\x12\x10\n\x08\x61llFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x0f\n\x07pathlen\x18\x05 \x01(\x05\x12\x1b\n\x02\x64\x63\x18\x06 \x03(\x0b\x32\x0f.CommDataCouple\"J\n\x10NavBorderDataGet\x12\r\n\x05jobId\x18\x01 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x02 \x01(\x05\x12\x11\n\tborderLen\x18\x03 \x01(\x05\"\x91\x01\n\x0cNavOptLineUp\x12\x12\n\nstartJobRI\x18\x01 \x01(\x05\x12\x10\n\x08\x65ndJobRI\x18\x02 \x01(\x05\x12\x10\n\x08\x61llFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x16\n\x0e\x63hannelDataLen\x18\x05 \x01(\x05\x12\x1b\n\x02\x64\x63\x18\x06 \x03(\x0b\x32\x0f.CommDataCouple\"~\n\x11NavOptiBorderInfo\x12\r\n\x05jobId\x18\x01 \x01(\x05\x12\x10\n\x08\x61llFrame\x18\x02 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x03 \x01(\x05\x12\x15\n\rborderDataLen\x18\x04 \x01(\x05\x12\x1b\n\x02\x64\x63\x18\x05 \x03(\x0b\x32\x0f.CommDataCouple\"\x81\x01\n\rNavOptObsInfo\x12\x12\n\nobstacleId\x18\x01 \x01(\x05\x12\x10\n\x08\x61llFrame\x18\x02 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x03 \x01(\x05\x12\x17\n\x0fobstacleDataLen\x18\x04 \x01(\x05\x12\x1b\n\x02\x64\x63\x18\x05 \x03(\x0b\x32\x0f.CommDataCouple\"\xb4\x01\n\x0bNavStartJob\x12\r\n\x05jobId\x18\x01 \x01(\x03\x12\x0e\n\x06jobVer\x18\x02 \x01(\x05\x12\x0f\n\x07jobMode\x18\x03 \x01(\x05\x12\x13\n\x0brainTactics\x18\x04 \x01(\x05\x12\x13\n\x0bknifeHeight\x18\x05 \x01(\x05\x12\r\n\x05speed\x18\x06 \x01(\x02\x12\x14\n\x0c\x63hannelWidth\x18\x07 \x01(\x05\x12\x11\n\tUltraWave\x18\x08 \x01(\x05\x12\x13\n\x0b\x63hannelMode\x18\t \x01(\x05\"\'\n\x0fNavTaskProgress\x12\x14\n\x0ctaskProgress\x18\x01 \x01(\x05\"\x1e\n\x0bNavResFrame\x12\x0f\n\x07\x66rameid\x18\x01 \x01(\x05\"|\n\x0eNavGetHashList\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x12\n\ntotalFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x05 \x01(\x06\x12\x10\n\x08reserved\x18\x06 \x01(\t\"\xb4\x01\n\x11NavGetHashListAck\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x12\n\ntotalFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x05 \x01(\x06\x12\x0f\n\x07hashLen\x18\x06 \x01(\x05\x12\x10\n\x08reserved\x18\x07 \x01(\t\x12\x0e\n\x06result\x18\x08 \x01(\x05\x12\x12\n\ndataCouple\x18\r \x03(\x03\"\xd6\x01\n\x0eNavGetCommData\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x03 \x01(\x05\x12\x0c\n\x04type\x18\x04 \x01(\x05\x12\x0c\n\x04hash\x18\x05 \x01(\x03\x12\x15\n\rpaternalHashA\x18\x06 \x01(\x03\x12\x15\n\rpaternalHashB\x18\x07 \x01(\x03\x12\x12\n\ntotalFrame\x18\x08 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\t \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\n \x01(\x06\x12\x10\n\x08reserved\x18\x0b \x01(\t\"H\n\x0eNavGetNameTime\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x02 \x01(\x06\x12\x13\n\x0bmodify_time\x18\x03 \x01(\x06\"\xc2\x02\n\x11NavGetCommDataAck\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x0e\n\x06result\x18\x03 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x04 \x01(\x05\x12\x0c\n\x04type\x18\x05 \x01(\x05\x12\x0c\n\x04Hash\x18\x06 \x01(\x06\x12\x15\n\rpaternalHashA\x18\x07 \x01(\x06\x12\x15\n\rpaternalHashB\x18\x08 \x01(\x06\x12\x12\n\ntotalFrame\x18\t \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\n \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x0b \x01(\x06\x12\x0f\n\x07\x64\x61taLen\x18\x0c \x01(\x05\x12#\n\ndataCouple\x18\r \x03(\x0b\x32\x0f.CommDataCouple\x12\x10\n\x08reserved\x18\x0e \x01(\t\x12!\n\x08nameTime\x18\x0f \x01(\x0b\x32\x0f.NavGetNameTime\"\xde\x02\n\x0fNavReqCoverPath\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\r\n\x05jobId\x18\x02 \x01(\x06\x12\x0e\n\x06jobVer\x18\x03 \x01(\x05\x12\x0f\n\x07jobMode\x18\x04 \x01(\x05\x12\x0e\n\x06subCmd\x18\x05 \x01(\x05\x12\x10\n\x08\x65\x64geMode\x18\x06 \x01(\x05\x12\x13\n\x0bknifeHeight\x18\x07 \x01(\x05\x12\x14\n\x0c\x63hannelWidth\x18\x08 \x01(\x05\x12\x11\n\tUltraWave\x18\t \x01(\x05\x12\x13\n\x0b\x63hannelMode\x18\n \x01(\x05\x12\x0e\n\x06toward\x18\x0b \x01(\x05\x12\r\n\x05speed\x18\x0c \x01(\x02\x12\x11\n\tzoneHashs\x18\r \x03(\x06\x12\x10\n\x08pathHash\x18\x0e \x01(\x06\x12\x10\n\x08reserved\x18\x0f \x01(\t\x12\x0e\n\x06result\x18\x10 \x01(\x05\x12\x13\n\x0btoward_mode\x18\x11 \x01(\x05\x12\x1d\n\x15toward_included_angle\x18\x12 \x01(\x05\"\xa7\x03\n\x15NavUploadZigZagResult\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\r\n\x05jobId\x18\x02 \x01(\x06\x12\x0e\n\x06jobVer\x18\x03 \x01(\x05\x12\x0e\n\x06result\x18\x04 \x01(\x05\x12\x0c\n\x04\x61rea\x18\x05 \x01(\x05\x12\x0c\n\x04time\x18\x06 \x01(\x05\x12\x14\n\x0ctotalZoneNum\x18\x07 \x01(\x05\x12\x1a\n\x12\x63urrentZonePathNum\x18\x08 \x01(\x05\x12\x19\n\x11\x63urrentZonePathId\x18\t \x01(\x05\x12\x13\n\x0b\x63urrentZone\x18\n \x01(\x05\x12\x13\n\x0b\x63urrentHash\x18\x0b \x01(\x06\x12\x12\n\ntotalFrame\x18\x0c \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\r \x01(\x05\x12\x13\n\x0b\x63hannelMode\x18\x0e \x01(\x05\x12\x15\n\rchannelModeId\x18\x0f \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x10 \x01(\x06\x12\x0f\n\x07\x64\x61taLen\x18\x11 \x01(\x05\x12\x10\n\x08reserved\x18\x12 \x01(\t\x12#\n\ndataCouple\x18\x13 \x03(\x0b\x32\x0f.CommDataCouple\x12\x0e\n\x06subCmd\x18\x14 \x01(\x05\"\xb0\x01\n\x18NavUploadZigZagResultAck\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x13\n\x0b\x63urrentZone\x18\x02 \x01(\x05\x12\x13\n\x0b\x63urrentHash\x18\x03 \x01(\x06\x12\x12\n\ntotalFrame\x18\x04 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x05 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x06 \x01(\x06\x12\x10\n\x08reserved\x18\x07 \x01(\t\x12\x0e\n\x06subCmd\x18\x08 \x01(\x05\"M\n\x0bNavTaskCtrl\x12\x0c\n\x04type\x18\x01 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x02 \x01(\x05\x12\x0e\n\x06result\x18\x03 \x01(\x05\x12\x10\n\x08reserved\x18\x04 \x01(\t\"o\n\x0bNavTaskIdRw\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x10\n\x08taskName\x18\x03 \x01(\t\x12\x0e\n\x06taskId\x18\x04 \x01(\t\x12\x0e\n\x06result\x18\x05 \x01(\x05\x12\x10\n\x08reserved\x18\x06 \x01(\t\"J\n\x12NavSysHashOverview\x12\x1a\n\x12\x63ommonhashOverview\x18\x01 \x01(\x06\x12\x18\n\x10pathHashOverview\x18\x02 \x01(\x06\"i\n\x11NavTaskBreakPoint\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\x0e\n\x06toward\x18\x03 \x01(\x05\x12\x0c\n\x04\x66lag\x18\x04 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x05 \x01(\x05\x12\x10\n\x08zoneHash\x18\x06 \x01(\x06\"\xc2\x05\n\rNavPlanJobSet\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x0c\n\x04\x61rea\x18\x03 \x01(\x05\x12\x10\n\x08workTime\x18\x04 \x01(\x05\x12\x0f\n\x07version\x18\x05 \x01(\t\x12\n\n\x02id\x18\x06 \x01(\t\x12\x0e\n\x06userId\x18\x07 \x01(\t\x12\x10\n\x08\x64\x65viceId\x18\x08 \x01(\t\x12\x0e\n\x06planId\x18\t \x01(\t\x12\x0e\n\x06taskId\x18\n \x01(\t\x12\r\n\x05jobId\x18\x0b \x01(\t\x12\x11\n\tstartTime\x18\x0c \x01(\t\x12\x0f\n\x07\x65ndTime\x18\r \x01(\t\x12\x0c\n\x04week\x18\x0e \x01(\x05\x12\x13\n\x0bknifeHeight\x18\x0f \x01(\x05\x12\r\n\x05model\x18\x10 \x01(\x05\x12\x10\n\x08\x65\x64geMode\x18\x11 \x01(\x05\x12\x14\n\x0crequiredTime\x18\x12 \x01(\x05\x12\x12\n\nrouteAngle\x18\x13 \x01(\x05\x12\x12\n\nrouteModel\x18\x14 \x01(\x05\x12\x14\n\x0crouteSpacing\x18\x15 \x01(\x05\x12\x19\n\x11ultrasonicBarrier\x18\x16 \x01(\x05\x12\x14\n\x0ctotalPlanNum\x18\x17 \x01(\x05\x12\x11\n\tPlanIndex\x18\x18 \x01(\x05\x12\x0e\n\x06result\x18\x19 \x01(\x05\x12\r\n\x05speed\x18\x1a \x01(\x02\x12\x10\n\x08taskName\x18\x1b \x01(\t\x12\x0f\n\x07jobName\x18\x1c \x01(\t\x12\x11\n\tzoneHashs\x18\x1d \x03(\x06\x12\x10\n\x08reserved\x18\x1e \x01(\t\x12\x11\n\tstartDate\x18\x1f \x01(\t\x12\x0f\n\x07\x65ndDate\x18  \x01(\t\x12\x13\n\x0btriggerType\x18! \x01(\x05\x12\x0b\n\x03\x64\x61y\x18\" \x01(\x05\x12\r\n\x05weeks\x18# \x03(\x07\x12\x18\n\x10remained_seconds\x18$ \x01(\x03\x12\x12\n\ntowardMode\x18% \x01(\x05\x12\x1b\n\x13towardIncludedAngle\x18& \x01(\x05\"\x97\x01\n\x10NavUnableTimeSet\x12\x0e\n\x06subCmd\x18\x01 \x01(\x05\x12\x10\n\x08\x64\x65viceId\x18\x02 \x01(\t\x12\x17\n\x0funableStartTime\x18\x03 \x01(\t\x12\x15\n\runableEndTime\x18\x04 \x01(\t\x12\x0e\n\x06result\x18\x05 \x01(\x05\x12\x10\n\x08reserved\x18\x06 \x01(\t\x12\x0f\n\x07trigger\x18\x07 \x01(\x05\"6\n\x0e\x63hargePileType\x12\x0e\n\x06toward\x18\x01 \x01(\x05\x12\t\n\x01x\x18\x02 \x01(\x02\x12\t\n\x01y\x18\x03 \x01(\x02\"J\n\x11SimulationCmdData\x12\x0e\n\x06subCmd\x18\x01 \x01(\x05\x12\x10\n\x08param_id\x18\x02 \x01(\x05\x12\x13\n\x0bparam_value\x18\x03 \x03(\x05\"%\n\x13WorkReportUpdateCmd\x12\x0e\n\x06subCmd\x18\x01 \x01(\x05\"<\n\x13WorkReportUpdateAck\x12\x13\n\x0bupdate_flag\x18\x01 \x01(\x08\x12\x10\n\x08info_num\x18\x02 \x01(\x05\"7\n\x11WorkReportCmdData\x12\x0e\n\x06subCmd\x18\x01 \x01(\x05\x12\x12\n\ngetInfoNum\x18\x02 \x01(\x05\"\xa3\x02\n\x11WorkReportInfoAck\x12\x16\n\x0einterrupt_flag\x18\x01 \x01(\x08\x12\x17\n\x0fstart_work_time\x18\x02 \x01(\x03\x12\x15\n\rend_work_time\x18\x03 \x01(\x03\x12\x16\n\x0ework_time_used\x18\x04 \x01(\x05\x12\x11\n\twork_ares\x18\x05 \x01(\x01\x12\x15\n\rwork_progress\x18\x06 \x01(\x05\x12\x17\n\x0fheight_of_knife\x18\x07 \x01(\x05\x12\x11\n\twork_type\x18\x08 \x01(\x05\x12\x13\n\x0bwork_result\x18\t \x01(\x05\x12\x15\n\rtotal_ack_num\x18\n \x01(\x05\x12\x17\n\x0f\x63urrent_ack_num\x18\x0b \x01(\x05\x12\x13\n\x0bjob_content\x18\x0c \x01(\x05\"\xb2\x01\n\x19\x61pp_request_cover_paths_t\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x12\n\ntotalFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x05 \x01(\x06\x12\x16\n\x0etransaction_id\x18\x06 \x01(\x03\x12\x10\n\x08reserved\x18\x07 \x03(\x03\x12\x11\n\thash_list\x18\x08 \x03(\x06\"\x99\x01\n\x13\x63over_path_packet_t\x12\x11\n\tpath_hash\x18\x01 \x01(\x06\x12\x11\n\tpath_type\x18\x02 \x01(\x05\x12\x12\n\npath_total\x18\x03 \x01(\x05\x12\x10\n\x08path_cur\x18\x04 \x01(\x05\x12\x11\n\tzone_hash\x18\x05 \x01(\x06\x12#\n\ndataCouple\x18\x06 \x03(\x0b\x32\x0f.CommDataCouple\"\xb2\x02\n\x13\x63over_path_upload_t\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06result\x18\x02 \x01(\x05\x12\x0e\n\x06subCmd\x18\x03 \x01(\x05\x12\x0c\n\x04\x61rea\x18\x04 \x01(\x05\x12\x0c\n\x04time\x18\x05 \x01(\x05\x12\x12\n\ntotalFrame\x18\x06 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x07 \x01(\x05\x12\x16\n\x0etotal_path_num\x18\x08 \x01(\x05\x12\x16\n\x0evalid_path_num\x18\t \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\n \x01(\x06\x12\x16\n\x0etransaction_id\x18\x0b \x01(\x03\x12\x10\n\x08reserved\x18\x0c \x03(\x03\x12\x0f\n\x07\x64\x61taLen\x18\r \x01(\x05\x12*\n\x0cpath_packets\x18\x0e \x03(\x0b\x32\x14.cover_path_packet_t\"M\n\x14zone_start_precent_t\x12\x10\n\x08\x64\x61taHash\x18\x01 \x01(\x06\x12\t\n\x01x\x18\x02 \x01(\x02\x12\t\n\x01y\x18\x03 \x01(\x02\x12\r\n\x05index\x18\x04 \x01(\x05\",\n\x0fvision_ctrl_msg\x12\x0c\n\x04type\x18\x01 \x01(\x05\x12\x0b\n\x03\x63md\x18\x02 \x01(\x05\"<\n\x11nav_sys_param_msg\x12\n\n\x02rw\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontext\x18\x03 \x01(\x05\"Q\n\x15nav_plan_task_execute\x12\x0e\n\x06subCmd\x18\x01 \x01(\x05\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06result\x18\x04 \x01(\x05\"y\n\tcostmap_t\x12\r\n\x05width\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x12\x10\n\x08\x63\x65nter_x\x18\x03 \x01(\x02\x12\x10\n\x08\x63\x65nter_y\x18\x04 \x01(\x02\x12\x0b\n\x03yaw\x18\x05 \x01(\x02\x12\x0b\n\x03res\x18\x06 \x01(\x02\x12\x0f\n\x07\x63ostmap\x18\x07 \x03(\x05\"\x94\x02\n\rsvg_message_t\x12\x0e\n\x06x_move\x18\x01 \x01(\x01\x12\x0e\n\x06y_move\x18\x02 \x01(\x01\x12\r\n\x05scale\x18\x03 \x01(\x01\x12\x0e\n\x06rotate\x18\x04 \x01(\x01\x12\x14\n\x0c\x62\x61se_width_m\x18\x05 \x01(\x01\x12\x15\n\rbase_height_m\x18\x06 \x01(\x01\x12\x16\n\x0e\x62\x61se_width_pix\x18\x07 \x01(\x05\x12\x17\n\x0f\x62\x61se_height_pix\x18\x08 \x01(\x05\x12\x15\n\rsvg_file_name\x18\t \x01(\t\x12\x15\n\rsvg_file_data\x18\n \x01(\t\x12\x12\n\nname_count\x18\x0b \x01(\x05\x12\x12\n\ndata_count\x18\x0c \x01(\x05\x12\x10\n\x08hide_svg\x18\r \x01(\x08\"\xc7\x01\n\x11svg_message_ack_t\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0e\n\x06subCmd\x18\x02 \x01(\x05\x12\x12\n\ntotalFrame\x18\x03 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x04 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x05 \x01(\x06\x12\x15\n\rpaternalHashA\x18\x06 \x01(\x06\x12\x0c\n\x04type\x18\x07 \x01(\x05\x12\x0e\n\x06result\x18\x08 \x01(\x05\x12#\n\x0bsvg_message\x18\t \x01(\x0b\x32\x0e.svg_message_t\"/\n\x13plan_task_name_id_t\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"<\n\x15nav_get_all_plan_task\x12#\n\x05tasks\x18\x01 \x03(\x0b\x32\x14.plan_task_name_id_t\",\n\x0e\x61rea_hash_name\x12\x0c\n\x04hash\x18\x01 \x01(\x06\x12\x0c\n\x04name\x18\x02 \x01(\t\"M\n\x15\x41ppGetAllAreaHashName\x12\x10\n\x08\x64\x65viceId\x18\x01 \x01(\t\x12\"\n\thashnames\x18\x02 \x03(\x0b\x32\x0f.area_hash_name\"c\n\x0eNavTaskCtrlAck\x12\x0c\n\x04type\x18\x01 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x02 \x01(\x05\x12\x0e\n\x06result\x18\x03 \x01(\x05\x12\x11\n\tnav_state\x18\x04 \x01(\x05\x12\x10\n\x08reserved\x18\x05 \x01(\t\"Y\n\rNavMapNameMsg\x12\n\n\x02rw\x18\x01 \x01(\x05\x12\x0c\n\x04hash\x18\x02 \x01(\x06\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06result\x18\x04 \x01(\x05\x12\x10\n\x08\x64\x65viceId\x18\x05 \x01(\t\"\x9a\x01\n\rNavEdgePoints\x12\x0e\n\x06result\x18\x01 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x02 \x01(\x05\x12\x0c\n\x04type\x18\x03 \x01(\x05\x12\x0c\n\x04Hash\x18\x04 \x01(\x06\x12\x12\n\ntotalFrame\x18\x05 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x06 \x01(\x05\x12#\n\ndataCouple\x18\x07 \x03(\x0b\x32\x0f.CommDataCouple\"x\n\x10NavEdgePointsAck\x12\x0e\n\x06result\x18\x01 \x01(\x05\x12\x0e\n\x06\x61\x63tion\x18\x02 \x01(\x05\x12\x0c\n\x04type\x18\x03 \x01(\x05\x12\x0c\n\x04Hash\x18\x04 \x01(\x06\x12\x12\n\ntotalFrame\x18\x05 \x01(\x05\x12\x14\n\x0c\x63urrentFrame\x18\x06 \x01(\x05\"\x8f\x02\n\x14ManualElementMessage\x12\x0c\n\x04pver\x18\x01 \x01(\x05\x12\x0c\n\x04type\x18\x02 \x01(\x05\x12\r\n\x05shape\x18\x03 \x01(\x05\x12\x0e\n\x06result\x18\x04 \x01(\x05\x12\x0e\n\x06subCmd\x18\x05 \x01(\x05\x12\x10\n\x08\x64\x61taHash\x18\x06 \x01(\x06\x12\x0e\n\x06ifHide\x18\x07 \x01(\x08\x12\x10\n\x08reserved\x18\x08 \x01(\t\x12\x17\n\x0fpoint1_center_x\x18\t \x01(\x01\x12\x17\n\x0fpoint1_center_y\x18\n \x01(\x01\x12\x16\n\x0epoint2_width_x\x18\x0b \x01(\x01\x12\x17\n\x0fpoint2_height_y\x18\x0c \x01(\x01\x12\x15\n\rrotate_radius\x18\r \x01(\x01\"\x81\x16\n\x07MctlNav\x12$\n\x0ctoapp_lat_up\x18\x01 \x01(\x0b\x32\x0c.NavLatLonUpH\x00\x12!\n\x0ctoapp_pos_up\x18\x02 \x01(\x0b\x32\t.NavPosUpH\x00\x12.\n\x13todev_chl_line_data\x18\x03 \x01(\x0b\x32\x0f.NavCHlLineDataH\x00\x12\'\n\x0ftoapp_task_info\x18\x04 \x01(\x0b\x32\x0c.NavTaskInfoH\x00\x12*\n\x11toapp_opt_line_up\x18\x05 \x01(\x0b\x32\r.NavOptLineUpH\x00\x12\x33\n\x15toapp_opt_border_info\x18\x06 \x01(\x0b\x32\x12.NavOptiBorderInfoH\x00\x12,\n\x12toapp_opt_obs_info\x18\x07 \x01(\x0b\x32\x0e.NavOptObsInfoH\x00\x12+\n\x13todev_task_info_ack\x18\x08 \x01(\x0b\x32\x0c.NavResFrameH\x00\x12\x31\n\x19todev_opt_border_info_ack\x18\t \x01(\x0b\x32\x0c.NavResFrameH\x00\x12.\n\x16todev_opt_obs_info_ack\x18\n \x01(\x0b\x32\x0c.NavResFrameH\x00\x12-\n\x15todev_opt_line_up_ack\x18\x0b \x01(\x0b\x32\x0c.NavResFrameH\x00\x12*\n\x0ftoapp_chgpileto\x18\x0c \x01(\x0b\x32\x0f.chargePileTypeH\x00\x12\x17\n\rtodev_sustask\x18\r \x01(\x05H\x00\x12\x18\n\x0etodev_rechgcmd\x18\x0e \x01(\x05H\x00\x12\x17\n\rtodev_edgecmd\x18\x0f \x01(\x05H\x00\x12\x1b\n\x11todev_draw_border\x18\x10 \x01(\x05H\x00\x12\x1f\n\x15todev_draw_border_end\x18\x11 \x01(\x05H\x00\x12\x18\n\x0etodev_draw_obs\x18\x12 \x01(\x05H\x00\x12\x1c\n\x12todev_draw_obs_end\x18\x13 \x01(\x05H\x00\x12\x18\n\x0etodev_chl_line\x18\x14 \x01(\x05H\x00\x12\x1c\n\x12todev_chl_line_end\x18\x15 \x01(\x05H\x00\x12\x19\n\x0ftodev_save_task\x18\x16 \x01(\x05H\x00\x12\x1d\n\x13todev_cancel_suscmd\x18\x17 \x01(\x05H\x00\x12\x1e\n\x14todev_reset_chg_pile\x18\x18 \x01(\x05H\x00\x12\x1f\n\x15todev_cancel_draw_cmd\x18\x19 \x01(\x05H\x00\x12$\n\x1atodev_one_touch_leave_pile\x18\x1a \x01(\x05H\x00\x12&\n\x0etodev_mow_task\x18\x1b \x01(\x0b\x32\x0c.NavStartJobH\x00\x12\'\n\x0ctoapp_bstate\x18\x1c \x01(\x0b\x32\x0f.NavBorderStateH\x00\x12\x1a\n\x10todev_lat_up_ack\x18\x1d \x01(\x05H\x00\x12(\n\rtodev_gethash\x18\x1e \x01(\x0b\x32\x0f.NavGetHashListH\x00\x12/\n\x11toapp_gethash_ack\x18\x1f \x01(\x0b\x32\x12.NavGetHashListAckH\x00\x12/\n\x14todev_get_commondata\x18  \x01(\x0b\x32\x0f.NavGetCommDataH\x00\x12\x36\n\x18toapp_get_commondata_ack\x18! \x01(\x0b\x32\x12.NavGetCommDataAckH\x00\x12\x31\n\x15\x62idire_reqconver_path\x18\" \x01(\x0b\x32\x10.NavReqCoverPathH\x00\x12.\n\x0ctoapp_zigzag\x18# \x01(\x0b\x32\x16.NavUploadZigZagResultH\x00\x12\x35\n\x10todev_zigzag_ack\x18$ \x01(\x0b\x32\x19.NavUploadZigZagResultAckH\x00\x12&\n\x0etodev_taskctrl\x18% \x01(\x0b\x32\x0c.NavTaskCtrlH\x00\x12%\n\rbidire_taskid\x18& \x01(\x0b\x32\x0c.NavTaskIdRwH\x00\x12&\n\x08toapp_bp\x18\' \x01(\x0b\x32\x12.NavTaskBreakPointH\x00\x12+\n\x11todev_planjob_set\x18( \x01(\x0b\x32\x0e.NavPlanJobSetH\x00\x12\x32\n\x15todev_unable_time_set\x18) \x01(\x0b\x32\x11.NavUnableTimeSetH\x00\x12,\n\x0esimulation_cmd\x18* \x01(\x0b\x32\x12.SimulationCmdDataH\x00\x12<\n\x1ctodev_work_report_update_cmd\x18+ \x01(\x0b\x32\x14.WorkReportUpdateCmdH\x00\x12<\n\x1ctoapp_work_report_update_ack\x18, \x01(\x0b\x32\x14.WorkReportUpdateAckH\x00\x12\x33\n\x15todev_work_report_cmd\x18- \x01(\x0b\x32\x12.WorkReportCmdDataH\x00\x12\x33\n\x15toapp_work_report_ack\x18. \x01(\x0b\x32\x12.WorkReportInfoAckH\x00\x12\x36\n\x18toapp_work_report_upload\x18/ \x01(\x0b\x32\x12.WorkReportInfoAckH\x00\x12=\n\x17\x61pp_request_cover_paths\x18\x30 \x01(\x0b\x32\x1a.app_request_cover_paths_tH\x00\x12\x31\n\x11\x63over_path_upload\x18\x31 \x01(\x0b\x32\x14.cover_path_upload_tH\x00\x12\x33\n\x12zone_start_precent\x18\x32 \x01(\x0b\x32\x15.zone_start_precent_tH\x00\x12\'\n\x0bvision_ctrl\x18\x33 \x01(\x0b\x32\x10.vision_ctrl_msgH\x00\x12/\n\x11nav_sys_param_cmd\x18\x34 \x01(\x0b\x32\x12.nav_sys_param_msgH\x00\x12\x33\n\x11plan_task_execute\x18\x35 \x01(\x0b\x32\x16.nav_plan_task_executeH\x00\x12#\n\rtoapp_costmap\x18\x36 \x01(\x0b\x32\n.costmap_tH\x00\x12\x31\n\x11plan_task_name_id\x18\x37 \x01(\x0b\x32\x14.plan_task_name_id_tH\x00\x12/\n\rall_plan_task\x18\x38 \x01(\x0b\x32\x16.nav_get_all_plan_taskH\x00\x12-\n\x12todev_taskctrl_ack\x18\x39 \x01(\x0b\x32\x0f.NavTaskCtrlAckH\x00\x12,\n\x12toapp_map_name_msg\x18: \x01(\x0b\x32\x0e.NavMapNameMsgH\x00\x12+\n\rtodev_svg_msg\x18; \x01(\x0b\x32\x12.svg_message_ack_tH\x00\x12+\n\rtoapp_svg_msg\x18< \x01(\x0b\x32\x12.svg_message_ack_tH\x00\x12\x35\n\x13toapp_all_hash_name\x18= \x01(\x0b\x32\x16.AppGetAllAreaHashNameH\x00\x12+\n\x11toapp_edge_points\x18> \x01(\x0b\x32\x0e.NavEdgePointsH\x00\x12\x32\n\x15toapp_edge_points_ack\x18? \x01(\x0b\x32\x11.NavEdgePointsAckH\x00\x12\x35\n\x14toapp_manual_element\x18@ \x01(\x0b\x32\x15.ManualElementMessageH\x00\x42\x0b\n\tSubNavMsgb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pymammotion.proto.mctrl_nav_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _NAVLATLONUP._serialized_start=69
  _NAVLATLONUP._serialized_end=108
  _NAVBORDERSTATE._serialized_start=110
  _NAVBORDERSTATE._serialized_end=143
  _NAVPOSUP._serialized_start=146
  _NAVPOSUP._serialized_end=347
  _NAVBORDERDATAGETACK._serialized_start=349
  _NAVBORDERDATAGETACK._serialized_end=407
  _NAVOBSTIBORDERDATAGET._serialized_start=409
  _NAVOBSTIBORDERDATAGET._serialized_end=499
  _NAVOBSTIBORDERDATAGETACK._serialized_start=501
  _NAVOBSTIBORDERDATAGETACK._serialized_end=572
  _NAVCHLLINEDATA._serialized_start=574
  _NAVCHLLINEDATA._serialized_end=674
  _NAVCHLLINEDATAACK._serialized_start=676
  _NAVCHLLINEDATAACK._serialized_end=755
  _NAVTASKINFO._serialized_start=757
  _NAVTASKINFO._serialized_end=884
  _NAVBORDERDATAGET._serialized_start=886
  _NAVBORDERDATAGET._serialized_end=960
  _NAVOPTLINEUP._serialized_start=963
  _NAVOPTLINEUP._serialized_end=1108
  _NAVOPTIBORDERINFO._serialized_start=1110
  _NAVOPTIBORDERINFO._serialized_end=1236
  _NAVOPTOBSINFO._serialized_start=1239
  _NAVOPTOBSINFO._serialized_end=1368
  _NAVSTARTJOB._serialized_start=1371
  _NAVSTARTJOB._serialized_end=1551
  _NAVTASKPROGRESS._serialized_start=1553
  _NAVTASKPROGRESS._serialized_end=1592
  _NAVRESFRAME._serialized_start=1594
  _NAVRESFRAME._serialized_end=1624
  _NAVGETHASHLIST._serialized_start=1626
  _NAVGETHASHLIST._serialized_end=1750
  _NAVGETHASHLISTACK._serialized_start=1753
  _NAVGETHASHLISTACK._serialized_end=1933
  _NAVGETCOMMDATA._serialized_start=1936
  _NAVGETCOMMDATA._serialized_end=2150
  _NAVGETNAMETIME._serialized_start=2152
  _NAVGETNAMETIME._serialized_end=2224
  _NAVGETCOMMDATAACK._serialized_start=2227
  _NAVGETCOMMDATAACK._serialized_end=2549
  _NAVREQCOVERPATH._serialized_start=2552
  _NAVREQCOVERPATH._serialized_end=2902
  _NAVUPLOADZIGZAGRESULT._serialized_start=2905
  _NAVUPLOADZIGZAGRESULT._serialized_end=3328
  _NAVUPLOADZIGZAGRESULTACK._serialized_start=3331
  _NAVUPLOADZIGZAGRESULTACK._serialized_end=3507
  _NAVTASKCTRL._serialized_start=3509
  _NAVTASKCTRL._serialized_end=3586
  _NAVTASKIDRW._serialized_start=3588
  _NAVTASKIDRW._serialized_end=3699
  _NAVSYSHASHOVERVIEW._serialized_start=3701
  _NAVSYSHASHOVERVIEW._serialized_end=3775
  _NAVTASKBREAKPOINT._serialized_start=3777
  _NAVTASKBREAKPOINT._serialized_end=3882
  _NAVPLANJOBSET._serialized_start=3885
  _NAVPLANJOBSET._serialized_end=4591
  _NAVUNABLETIMESET._serialized_start=4594
  _NAVUNABLETIMESET._serialized_end=4745
  _CHARGEPILETYPE._serialized_start=4747
  _CHARGEPILETYPE._serialized_end=4801
  _SIMULATIONCMDDATA._serialized_start=4803
  _SIMULATIONCMDDATA._serialized_end=4877
  _WORKREPORTUPDATECMD._serialized_start=4879
  _WORKREPORTUPDATECMD._serialized_end=4916
  _WORKREPORTUPDATEACK._serialized_start=4918
  _WORKREPORTUPDATEACK._serialized_end=4978
  _WORKREPORTCMDDATA._serialized_start=4980
  _WORKREPORTCMDDATA._serialized_end=5035
  _WORKREPORTINFOACK._serialized_start=5038
  _WORKREPORTINFOACK._serialized_end=5329
  _APP_REQUEST_COVER_PATHS_T._serialized_start=5332
  _APP_REQUEST_COVER_PATHS_T._serialized_end=5510
  _COVER_PATH_PACKET_T._serialized_start=5513
  _COVER_PATH_PACKET_T._serialized_end=5666
  _COVER_PATH_UPLOAD_T._serialized_start=5669
  _COVER_PATH_UPLOAD_T._serialized_end=5975
  _ZONE_START_PRECENT_T._serialized_start=5977
  _ZONE_START_PRECENT_T._serialized_end=6054
  _VISION_CTRL_MSG._serialized_start=6056
  _VISION_CTRL_MSG._serialized_end=6100
  _NAV_SYS_PARAM_MSG._serialized_start=6102
  _NAV_SYS_PARAM_MSG._serialized_end=6162
  _NAV_PLAN_TASK_EXECUTE._serialized_start=6164
  _NAV_PLAN_TASK_EXECUTE._serialized_end=6245
  _COSTMAP_T._serialized_start=6247
  _COSTMAP_T._serialized_end=6368
  _SVG_MESSAGE_T._serialized_start=6371
  _SVG_MESSAGE_T._serialized_end=6647
  _SVG_MESSAGE_ACK_T._serialized_start=6650
  _SVG_MESSAGE_ACK_T._serialized_end=6849
  _PLAN_TASK_NAME_ID_T._serialized_start=6851
  _PLAN_TASK_NAME_ID_T._serialized_end=6898
  _NAV_GET_ALL_PLAN_TASK._serialized_start=6900
  _NAV_GET_ALL_PLAN_TASK._serialized_end=6960
  _AREA_HASH_NAME._serialized_start=6962
  _AREA_HASH_NAME._serialized_end=7006
  _APPGETALLAREAHASHNAME._serialized_start=7008
  _APPGETALLAREAHASHNAME._serialized_end=7085
  _NAVTASKCTRLACK._serialized_start=7087
  _NAVTASKCTRLACK._serialized_end=7186
  _NAVMAPNAMEMSG._serialized_start=7188
  _NAVMAPNAMEMSG._serialized_end=7277
  _NAVEDGEPOINTS._serialized_start=7280
  _NAVEDGEPOINTS._serialized_end=7434
  _NAVEDGEPOINTSACK._serialized_start=7436
  _NAVEDGEPOINTSACK._serialized_end=7556
  _MANUALELEMENTMESSAGE._serialized_start=7559
  _MANUALELEMENTMESSAGE._serialized_end=7830
  _MCTLNAV._serialized_start=7833
  _MCTLNAV._serialized_end=10650
# @@protoc_insertion_point(module_scope)



import json

from pymammotion.aliyun.cloud_gateway import DeviceOfflineException, GatewayTimeoutException, SetupException
from pymammotion.data.model.device import RTKDevice
from pymammotion.http.model.http import CheckDeviceVersion
from pymammotion.mammotion.devices.mammotion import Mammotion


class HomeAssistantRTKApi:
    def __init__(self) -> None:
        self._mammotion = Mammotion()

    @property
    def mammotion(self) -> Mammotion:
        return self._mammotion

    async def update(self, device_name: str) -> RTKDevice:
        """Update RTK data."""
        device = self.mammotion.get_rtk_device_by_name(device_name)
        try:
            response = await device.cloud_client.get_device_properties(device.iot_id)
            if response.code == 200:
                if data := response.data:
                    if ota_progress := data.otaProgress:
                        device.state.update_check = CheckDeviceVersion.from_dict(ota_progress.value)
                    if network_info := data.networkInfo:
                        network = json.loads(network_info.value)
                        device.state.wifi_rssi = network["wifi_rssi"]
                        device.state.wifi_sta_mac = network["wifi_sta_mac"]
                        device.state.bt_mac = network["bt_mac"]
                    if coordinate := data.coordinate:
                        coord_val = json.loads(coordinate.value)
                        if device.state.lat == 0:
                            device.state.lat = coord_val["lat"]
                        if device.state.lon == 0:
                            device.state.lon = coord_val["lon"]
                    if device_version := data.deviceVersion:
                        device.state.device_version = device_version.value
            device.state.online = True

            ota_info = await device.cloud_client.mammotion_http.get_device_ota_firmware([device.state.iot_id])
            if check_versions := ota_info.data:
                for check_version in check_versions:
                    if check_version.device_id == device.state.iot_id:
                        device.state.update_check = check_version
            return device.state
        except SetupException:
            """Cloud IOT Gateway is not setup."""
            return device.state
        except DeviceOfflineException:
            device.state.online = False
        except GatewayTimeoutException:
            """Gateway is timing out again."""
        return device.state



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

APN_AUTH_CHAP: apn_auth_type
APN_AUTH_NONE: apn_auth_type
APN_AUTH_PAP: apn_auth_type
APN_AUTH_PAP_CHAP: apn_auth_type
DESCRIPTOR: _descriptor.FileDescriptor
DRV_RESULT_FAIL: DrvDevInfoResult
DRV_RESULT_NOTSUP: DrvDevInfoResult
DRV_RESULT_SUC: DrvDevInfoResult
DirectConnectWifi: WifiConfType
DisconnectWifi: WifiConfType
FILE_TYPE_ALL: DrvUploadFileFileType
FILE_TYPE_NAVLOG: DrvUploadFileFileType
FILE_TYPE_RTKLOG: DrvUploadFileFileType
FILE_TYPE_SYSLOG: DrvUploadFileFileType
ForgetWifi: WifiConfType
IOT_TYPE_OFFLINE: iot_conctrl_type
IOT_TYPE_ONLINE: iot_conctrl_type
IOT_TYPE_RESET: iot_conctrl_type
MNET_LINK_2G: mnet_link_type
MNET_LINK_3G: mnet_link_type
MNET_LINK_4G: mnet_link_type
MNET_LINK_5G: mnet_link_type
MNET_LINK_NONE: mnet_link_type
NET_TYPE_MNET: net_type
NET_TYPE_WIFI: net_type
ReconnectWifi: WifiConfType
SIM_INPUT_PIN: sim_card_sta
SIM_INPUT_PUK: sim_card_sta
SIM_INVALID: sim_card_sta
SIM_NONE: sim_card_sta
SIM_NO_CARD: sim_card_sta
SIM_OK: sim_card_sta
set_enable: WifiConfType

class BleLogUploadUpdateProgress(_message.Message):
    __slots__ = ["bizId", "fIdx", "fNum", "fPath", "fSize", "prog", "speed", "status", "subProg", "ver"]
    BIZID_FIELD_NUMBER: _ClassVar[int]
    FIDX_FIELD_NUMBER: _ClassVar[int]
    FNUM_FIELD_NUMBER: _ClassVar[int]
    FPATH_FIELD_NUMBER: _ClassVar[int]
    FSIZE_FIELD_NUMBER: _ClassVar[int]
    PROG_FIELD_NUMBER: _ClassVar[int]
    SPEED_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    SUBPROG_FIELD_NUMBER: _ClassVar[int]
    VER_FIELD_NUMBER: _ClassVar[int]
    bizId: str
    fIdx: int
    fNum: int
    fPath: str
    fSize: int
    prog: int
    speed: int
    status: int
    subProg: int
    ver: int
    def __init__(self, ver: _Optional[int] = ..., status: _Optional[int] = ..., bizId: _Optional[str] = ..., prog: _Optional[int] = ..., subProg: _Optional[int] = ..., fNum: _Optional[int] = ..., speed: _Optional[int] = ..., fIdx: _Optional[int] = ..., fPath: _Optional[str] = ..., fSize: _Optional[int] = ...) -> None: ...

class BleSignatureReq(_message.Message):
    __slots__ = ["random_data", "signature_data"]
    RANDOM_DATA_FIELD_NUMBER: _ClassVar[int]
    SIGNATURE_DATA_FIELD_NUMBER: _ClassVar[int]
    random_data: str
    signature_data: str
    def __init__(self, signature_data: _Optional[str] = ..., random_data: _Optional[str] = ...) -> None: ...

class BleTestBytes(_message.Message):
    __slots__ = ["data", "seqs"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    SEQS_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[int]
    seqs: int
    def __init__(self, seqs: _Optional[int] = ..., data: _Optional[_Iterable[int]] = ...) -> None: ...

class DevNet(_message.Message):
    __slots__ = ["bir_testdata", "toapp_ListUpload", "toapp_WifiConf", "toapp_WifiMsg", "toapp_devinfo_resp", "toapp_get_mnet_cfg_rsp", "toapp_log_upload_update_progress", "toapp_mnet_info_rsp", "toapp_networkinfo_rsp", "toapp_set_mnet_cfg_rsp", "toapp_upgrade_report", "toapp_uploadfile_rsp", "toapp_wifi_iot_status", "todev_ConfType", "todev_WifiListUpload", "todev_WifiMsgUpload", "todev_Wifi_Configuration", "todev_ble_sync", "todev_devinfo_req", "todev_get_mnet_cfg_req", "todev_log_data_cancel", "todev_mnet_info_req", "todev_networkinfo_req", "todev_req_log_info", "todev_set_ble_mtu", "todev_set_dds2zmq", "todev_set_iot_offline_req", "todev_set_mnet_cfg_req", "todev_uploadfile_req", "todev_verify_signature_req"]
    BIR_TESTDATA_FIELD_NUMBER: _ClassVar[int]
    TOAPP_DEVINFO_RESP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_GET_MNET_CFG_RSP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_LISTUPLOAD_FIELD_NUMBER: _ClassVar[int]
    TOAPP_LOG_UPLOAD_UPDATE_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    TOAPP_MNET_INFO_RSP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_NETWORKINFO_RSP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_SET_MNET_CFG_RSP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_UPGRADE_REPORT_FIELD_NUMBER: _ClassVar[int]
    TOAPP_UPLOADFILE_RSP_FIELD_NUMBER: _ClassVar[int]
    TOAPP_WIFICONF_FIELD_NUMBER: _ClassVar[int]
    TOAPP_WIFIMSG_FIELD_NUMBER: _ClassVar[int]
    TOAPP_WIFI_IOT_STATUS_FIELD_NUMBER: _ClassVar[int]
    TODEV_BLE_SYNC_FIELD_NUMBER: _ClassVar[int]
    TODEV_CONFTYPE_FIELD_NUMBER: _ClassVar[int]
    TODEV_DEVINFO_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_GET_MNET_CFG_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_LOG_DATA_CANCEL_FIELD_NUMBER: _ClassVar[int]
    TODEV_MNET_INFO_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_NETWORKINFO_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_REQ_LOG_INFO_FIELD_NUMBER: _ClassVar[int]
    TODEV_SET_BLE_MTU_FIELD_NUMBER: _ClassVar[int]
    TODEV_SET_DDS2ZMQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_SET_IOT_OFFLINE_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_SET_MNET_CFG_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_UPLOADFILE_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_VERIFY_SIGNATURE_REQ_FIELD_NUMBER: _ClassVar[int]
    TODEV_WIFILISTUPLOAD_FIELD_NUMBER: _ClassVar[int]
    TODEV_WIFIMSGUPLOAD_FIELD_NUMBER: _ClassVar[int]
    TODEV_WIFI_CONFIGURATION_FIELD_NUMBER: _ClassVar[int]
    bir_testdata: BleTestBytes
    toapp_ListUpload: DrvListUpload
    toapp_WifiConf: DrvWifiConf
    toapp_WifiMsg: DrvWifiMsg
    toapp_devinfo_resp: DrvDevInfoResp
    toapp_get_mnet_cfg_rsp: GetMnetCfgRsp
    toapp_log_upload_update_progress: BleLogUploadUpdateProgress
    toapp_mnet_info_rsp: GetMnetInfoRsp
    toapp_networkinfo_rsp: GetNetworkInfoRsp
    toapp_set_mnet_cfg_rsp: SetMnetCfgRsp
    toapp_upgrade_report: DrvUpgradeReport
    toapp_uploadfile_rsp: DrvUploadFileToAppRsp
    toapp_wifi_iot_status: WifiIotStatusReport
    todev_ConfType: WifiConfType
    todev_WifiListUpload: DrvWifiList
    todev_WifiMsgUpload: DrvWifiUpload
    todev_Wifi_Configuration: DrvWifiSet
    todev_ble_sync: int
    todev_devinfo_req: DrvDevInfoReq
    todev_get_mnet_cfg_req: GetMnetCfgReq
    todev_log_data_cancel: DrvUploadFileCancel
    todev_mnet_info_req: GetMnetInfoReq
    todev_networkinfo_req: GetNetworkInfoReq
    todev_req_log_info: DrvUploadFileReq
    todev_set_ble_mtu: SetDrvBleMTU
    todev_set_dds2zmq: DrvDebugDdsZmq
    todev_set_iot_offline_req: iot_conctrl_type
    todev_set_mnet_cfg_req: SetMnetCfgReq
    todev_uploadfile_req: DrvUploadFileToAppReq
    todev_verify_signature_req: BleSignatureReq
    def __init__(self, todev_ble_sync: _Optional[int] = ..., todev_ConfType: _Optional[_Union[WifiConfType, str]] = ..., todev_WifiMsgUpload: _Optional[_Union[DrvWifiUpload, _Mapping]] = ..., todev_WifiListUpload: _Optional[_Union[DrvWifiList, _Mapping]] = ..., todev_Wifi_Configuration: _Optional[_Union[DrvWifiSet, _Mapping]] = ..., toapp_WifiMsg: _Optional[_Union[DrvWifiMsg, _Mapping]] = ..., toapp_WifiConf: _Optional[_Union[DrvWifiConf, _Mapping]] = ..., toapp_ListUpload: _Optional[_Union[DrvListUpload, _Mapping]] = ..., todev_req_log_info: _Optional[_Union[DrvUploadFileReq, _Mapping]] = ..., todev_log_data_cancel: _Optional[_Union[DrvUploadFileCancel, _Mapping]] = ..., todev_devinfo_req: _Optional[_Union[DrvDevInfoReq, _Mapping]] = ..., toapp_devinfo_resp: _Optional[_Union[DrvDevInfoResp, _Mapping]] = ..., toapp_upgrade_report: _Optional[_Union[DrvUpgradeReport, _Mapping]] = ..., toapp_wifi_iot_status: _Optional[_Union[WifiIotStatusReport, _Mapping]] = ..., todev_uploadfile_req: _Optional[_Union[DrvUploadFileToAppReq, _Mapping]] = ..., toapp_uploadfile_rsp: _Optional[_Union[DrvUploadFileToAppRsp, _Mapping]] = ..., todev_networkinfo_req: _Optional[_Union[GetNetworkInfoReq, _Mapping]] = ..., toapp_networkinfo_rsp: _Optional[_Union[GetNetworkInfoRsp, _Mapping]] = ..., bir_testdata: _Optional[_Union[BleTestBytes, _Mapping]] = ..., todev_mnet_info_req: _Optional[_Union[GetMnetInfoReq, _Mapping]] = ..., toapp_mnet_info_rsp: _Optional[_Union[GetMnetInfoRsp, _Mapping]] = ..., todev_get_mnet_cfg_req: _Optional[_Union[GetMnetCfgReq, _Mapping]] = ..., toapp_get_mnet_cfg_rsp: _Optional[_Union[GetMnetCfgRsp, _Mapping]] = ..., todev_set_mnet_cfg_req: _Optional[_Union[SetMnetCfgReq, _Mapping]] = ..., toapp_set_mnet_cfg_rsp: _Optional[_Union[SetMnetCfgRsp, _Mapping]] = ..., todev_set_dds2zmq: _Optional[_Union[DrvDebugDdsZmq, _Mapping]] = ..., todev_set_ble_mtu: _Optional[_Union[SetDrvBleMTU, _Mapping]] = ..., todev_set_iot_offline_req: _Optional[_Union[iot_conctrl_type, str]] = ..., todev_verify_signature_req: _Optional[_Union[BleSignatureReq, _Mapping]] = ..., toapp_log_upload_update_progress: _Optional[_Union[BleLogUploadUpdateProgress, _Mapping]] = ...) -> None: ...

class DrvDebugDdsZmq(_message.Message):
    __slots__ = ["is_enable", "rx_topic_name", "tx_zmq_url"]
    IS_ENABLE_FIELD_NUMBER: _ClassVar[int]
    RX_TOPIC_NAME_FIELD_NUMBER: _ClassVar[int]
    TX_ZMQ_URL_FIELD_NUMBER: _ClassVar[int]
    is_enable: bool
    rx_topic_name: str
    tx_zmq_url: str
    def __init__(self, is_enable: bool = ..., rx_topic_name: _Optional[str] = ..., tx_zmq_url: _Optional[str] = ...) -> None: ...

class DrvDevInfoReq(_message.Message):
    __slots__ = ["req_ids"]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    req_ids: _containers.RepeatedCompositeFieldContainer[DrvDevInfoReqId]
    def __init__(self, req_ids: _Optional[_Iterable[_Union[DrvDevInfoReqId, _Mapping]]] = ...) -> None: ...

class DrvDevInfoReqId(_message.Message):
    __slots__ = ["id", "type"]
    ID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    id: int
    type: int
    def __init__(self, id: _Optional[int] = ..., type: _Optional[int] = ...) -> None: ...

class DrvDevInfoResp(_message.Message):
    __slots__ = ["resp_ids"]
    RESP_IDS_FIELD_NUMBER: _ClassVar[int]
    resp_ids: _containers.RepeatedCompositeFieldContainer[DrvDevInfoRespId]
    def __init__(self, resp_ids: _Optional[_Iterable[_Union[DrvDevInfoRespId, _Mapping]]] = ...) -> None: ...

class DrvDevInfoRespId(_message.Message):
    __slots__ = ["id", "info", "res", "type"]
    ID_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    RES_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    id: int
    info: str
    res: DrvDevInfoResult
    type: int
    def __init__(self, id: _Optional[int] = ..., type: _Optional[int] = ..., res: _Optional[_Union[DrvDevInfoResult, str]] = ..., info: _Optional[str] = ...) -> None: ...

class DrvListUpload(_message.Message):
    __slots__ = ["Memssid", "current", "rssi", "status", "sum"]
    CURRENT_FIELD_NUMBER: _ClassVar[int]
    MEMSSID_FIELD_NUMBER: _ClassVar[int]
    Memssid: str
    RSSI_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    SUM_FIELD_NUMBER: _ClassVar[int]
    current: int
    rssi: int
    status: int
    sum: int
    def __init__(self, sum: _Optional[int] = ..., current: _Optional[int] = ..., status: _Optional[int] = ..., Memssid: _Optional[str] = ..., rssi: _Optional[int] = ...) -> None: ...

class DrvUpgradeReport(_message.Message):
    __slots__ = ["devname", "message", "otaid", "progress", "properties", "recv_cnt", "result", "version"]
    DEVNAME_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    OTAID_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    RECV_CNT_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    devname: str
    message: str
    otaid: str
    progress: int
    properties: str
    recv_cnt: int
    result: int
    version: str
    def __init__(self, devname: _Optional[str] = ..., otaid: _Optional[str] = ..., version: _Optional[str] = ..., progress: _Optional[int] = ..., result: _Optional[int] = ..., message: _Optional[str] = ..., properties: _Optional[str] = ..., recv_cnt: _Optional[int] = ...) -> None: ...

class DrvUploadFileCancel(_message.Message):
    __slots__ = ["bizId"]
    BIZID_FIELD_NUMBER: _ClassVar[int]
    bizId: str
    def __init__(self, bizId: _Optional[str] = ...) -> None: ...

class DrvUploadFileReq(_message.Message):
    __slots__ = ["bizId", "num", "type", "url", "userId"]
    BIZID_FIELD_NUMBER: _ClassVar[int]
    NUM_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    USERID_FIELD_NUMBER: _ClassVar[int]
    bizId: str
    num: int
    type: int
    url: str
    userId: str
    def __init__(self, bizId: _Optional[str] = ..., url: _Optional[str] = ..., userId: _Optional[str] = ..., num: _Optional[int] = ..., type: _Optional[int] = ...) -> None: ...

class DrvUploadFileToAppReq(_message.Message):
    __slots__ = ["bizId", "num", "operation", "serverIp", "serverPort", "type"]
    BIZID_FIELD_NUMBER: _ClassVar[int]
    NUM_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    SERVERIP_FIELD_NUMBER: _ClassVar[int]
    SERVERPORT_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    bizId: str
    num: int
    operation: int
    serverIp: int
    serverPort: int
    type: int
    def __init__(self, bizId: _Optional[str] = ..., operation: _Optional[int] = ..., serverIp: _Optional[int] = ..., serverPort: _Optional[int] = ..., num: _Optional[int] = ..., type: _Optional[int] = ...) -> None: ...

class DrvUploadFileToAppRsp(_message.Message):
    __slots__ = ["bizId", "operation", "result"]
    BIZID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    bizId: str
    operation: int
    result: int
    def __init__(self, bizId: _Optional[str] = ..., operation: _Optional[int] = ..., result: _Optional[int] = ...) -> None: ...

class DrvWifiConf(_message.Message):
    __slots__ = ["Confssid", "code", "succFlag"]
    CODE_FIELD_NUMBER: _ClassVar[int]
    CONFSSID_FIELD_NUMBER: _ClassVar[int]
    Confssid: str
    SUCCFLAG_FIELD_NUMBER: _ClassVar[int]
    code: int
    succFlag: bool
    def __init__(self, succFlag: bool = ..., code: _Optional[int] = ..., Confssid: _Optional[str] = ...) -> None: ...

class DrvWifiList(_message.Message):
    __slots__ = ["nvs_wifi_upload"]
    NVS_WIFI_UPLOAD_FIELD_NUMBER: _ClassVar[int]
    nvs_wifi_upload: int
    def __init__(self, nvs_wifi_upload: _Optional[int] = ...) -> None: ...

class DrvWifiMsg(_message.Message):
    __slots__ = ["devicename", "ip", "msgssid", "password", "productkey", "rssi", "status1", "status2", "wifi_enable"]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    IP_FIELD_NUMBER: _ClassVar[int]
    MSGSSID_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    PRODUCTKEY_FIELD_NUMBER: _ClassVar[int]
    RSSI_FIELD_NUMBER: _ClassVar[int]
    STATUS1_FIELD_NUMBER: _ClassVar[int]
    STATUS2_FIELD_NUMBER: _ClassVar[int]
    WIFI_ENABLE_FIELD_NUMBER: _ClassVar[int]
    devicename: str
    ip: str
    msgssid: str
    password: str
    productkey: str
    rssi: int
    status1: bool
    status2: bool
    wifi_enable: bool
    def __init__(self, status1: bool = ..., status2: bool = ..., ip: _Optional[str] = ..., msgssid: _Optional[str] = ..., password: _Optional[str] = ..., rssi: _Optional[int] = ..., productkey: _Optional[str] = ..., devicename: _Optional[str] = ..., wifi_enable: bool = ...) -> None: ...

class DrvWifiSet(_message.Message):
    __slots__ = ["Confssid", "configParam", "wifi_enable"]
    CONFIGPARAM_FIELD_NUMBER: _ClassVar[int]
    CONFSSID_FIELD_NUMBER: _ClassVar[int]
    Confssid: str
    WIFI_ENABLE_FIELD_NUMBER: _ClassVar[int]
    configParam: int
    wifi_enable: bool
    def __init__(self, configParam: _Optional[int] = ..., Confssid: _Optional[str] = ..., wifi_enable: bool = ...) -> None: ...

class DrvWifiUpload(_message.Message):
    __slots__ = ["wifi_msg_upload"]
    WIFI_MSG_UPLOAD_FIELD_NUMBER: _ClassVar[int]
    wifi_msg_upload: int
    def __init__(self, wifi_msg_upload: _Optional[int] = ...) -> None: ...

class GetMnetCfgReq(_message.Message):
    __slots__ = ["req_ids"]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    req_ids: int
    def __init__(self, req_ids: _Optional[int] = ...) -> None: ...

class GetMnetCfgRsp(_message.Message):
    __slots__ = ["cfg", "req_ids", "result"]
    CFG_FIELD_NUMBER: _ClassVar[int]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    cfg: MnetCfg
    req_ids: int
    result: int
    def __init__(self, req_ids: _Optional[int] = ..., result: _Optional[int] = ..., cfg: _Optional[_Union[MnetCfg, _Mapping]] = ...) -> None: ...

class GetMnetInfoReq(_message.Message):
    __slots__ = ["req_ids"]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    req_ids: int
    def __init__(self, req_ids: _Optional[int] = ...) -> None: ...

class GetMnetInfoRsp(_message.Message):
    __slots__ = ["mnet", "req_ids", "result"]
    MNET_FIELD_NUMBER: _ClassVar[int]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    mnet: MnetInfo
    req_ids: int
    result: int
    def __init__(self, req_ids: _Optional[int] = ..., result: _Optional[int] = ..., mnet: _Optional[_Union[MnetInfo, _Mapping]] = ...) -> None: ...

class GetNetworkInfoReq(_message.Message):
    __slots__ = ["req_ids"]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    req_ids: int
    def __init__(self, req_ids: _Optional[int] = ...) -> None: ...

class GetNetworkInfoRsp(_message.Message):
    __slots__ = ["gateway", "ip", "mask", "req_ids", "wifi_mac", "wifi_rssi", "wifi_ssid"]
    GATEWAY_FIELD_NUMBER: _ClassVar[int]
    IP_FIELD_NUMBER: _ClassVar[int]
    MASK_FIELD_NUMBER: _ClassVar[int]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    WIFI_MAC_FIELD_NUMBER: _ClassVar[int]
    WIFI_RSSI_FIELD_NUMBER: _ClassVar[int]
    WIFI_SSID_FIELD_NUMBER: _ClassVar[int]
    gateway: int
    ip: int
    mask: int
    req_ids: int
    wifi_mac: str
    wifi_rssi: int
    wifi_ssid: str
    def __init__(self, req_ids: _Optional[int] = ..., wifi_ssid: _Optional[str] = ..., wifi_mac: _Optional[str] = ..., wifi_rssi: _Optional[int] = ..., ip: _Optional[int] = ..., mask: _Optional[int] = ..., gateway: _Optional[int] = ...) -> None: ...

class MnetApn(_message.Message):
    __slots__ = ["apn_alias", "apn_name", "auth", "cid", "password", "username"]
    APN_ALIAS_FIELD_NUMBER: _ClassVar[int]
    APN_NAME_FIELD_NUMBER: _ClassVar[int]
    AUTH_FIELD_NUMBER: _ClassVar[int]
    CID_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    apn_alias: str
    apn_name: str
    auth: apn_auth_type
    cid: int
    password: str
    username: str
    def __init__(self, cid: _Optional[int] = ..., apn_alias: _Optional[str] = ..., apn_name: _Optional[str] = ..., auth: _Optional[_Union[apn_auth_type, str]] = ..., username: _Optional[str] = ..., password: _Optional[str] = ...) -> None: ...

class MnetApnCfg(_message.Message):
    __slots__ = ["apn", "apn_used_idx"]
    APN_FIELD_NUMBER: _ClassVar[int]
    APN_USED_IDX_FIELD_NUMBER: _ClassVar[int]
    apn: _containers.RepeatedCompositeFieldContainer[MnetApn]
    apn_used_idx: int
    def __init__(self, apn_used_idx: _Optional[int] = ..., apn: _Optional[_Iterable[_Union[MnetApn, _Mapping]]] = ...) -> None: ...

class MnetApnSetCfg(_message.Message):
    __slots__ = ["cfg", "use_default"]
    CFG_FIELD_NUMBER: _ClassVar[int]
    USE_DEFAULT_FIELD_NUMBER: _ClassVar[int]
    cfg: MnetApnCfg
    use_default: bool
    def __init__(self, use_default: bool = ..., cfg: _Optional[_Union[MnetApnCfg, _Mapping]] = ...) -> None: ...

class MnetCfg(_message.Message):
    __slots__ = ["apn", "auto_select", "inet_enable", "mnet_enable", "type"]
    APN_FIELD_NUMBER: _ClassVar[int]
    AUTO_SELECT_FIELD_NUMBER: _ClassVar[int]
    INET_ENABLE_FIELD_NUMBER: _ClassVar[int]
    MNET_ENABLE_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    apn: MnetApnSetCfg
    auto_select: bool
    inet_enable: bool
    mnet_enable: bool
    type: net_type
    def __init__(self, mnet_enable: bool = ..., inet_enable: bool = ..., type: _Optional[_Union[net_type, str]] = ..., apn: _Optional[_Union[MnetApnSetCfg, _Mapping]] = ..., auto_select: bool = ...) -> None: ...

class MnetInfo(_message.Message):
    __slots__ = ["iccid", "imei", "imsi", "inet", "link_type", "model", "operator", "revision", "rssi", "sim"]
    ICCID_FIELD_NUMBER: _ClassVar[int]
    IMEI_FIELD_NUMBER: _ClassVar[int]
    IMSI_FIELD_NUMBER: _ClassVar[int]
    INET_FIELD_NUMBER: _ClassVar[int]
    LINK_TYPE_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    OPERATOR_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    RSSI_FIELD_NUMBER: _ClassVar[int]
    SIM_FIELD_NUMBER: _ClassVar[int]
    iccid: str
    imei: str
    imsi: str
    inet: mnet_inet_status
    link_type: mnet_link_type
    model: str
    operator: str
    revision: str
    rssi: int
    sim: sim_card_sta
    def __init__(self, model: _Optional[str] = ..., revision: _Optional[str] = ..., imei: _Optional[str] = ..., sim: _Optional[_Union[sim_card_sta, str]] = ..., imsi: _Optional[str] = ..., link_type: _Optional[_Union[mnet_link_type, str]] = ..., rssi: _Optional[int] = ..., inet: _Optional[_Union[mnet_inet_status, _Mapping]] = ..., iccid: _Optional[str] = ..., operator: _Optional[str] = ...) -> None: ...

class SetDrvBleMTU(_message.Message):
    __slots__ = ["mtu_count"]
    MTU_COUNT_FIELD_NUMBER: _ClassVar[int]
    mtu_count: int
    def __init__(self, mtu_count: _Optional[int] = ...) -> None: ...

class SetMnetCfgReq(_message.Message):
    __slots__ = ["cfg", "req_ids"]
    CFG_FIELD_NUMBER: _ClassVar[int]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    cfg: MnetCfg
    req_ids: int
    def __init__(self, req_ids: _Optional[int] = ..., cfg: _Optional[_Union[MnetCfg, _Mapping]] = ...) -> None: ...

class SetMnetCfgRsp(_message.Message):
    __slots__ = ["req_ids", "result"]
    REQ_IDS_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    req_ids: int
    result: int
    def __init__(self, req_ids: _Optional[int] = ..., result: _Optional[int] = ...) -> None: ...

class WifiIotStatusReport(_message.Message):
    __slots__ = ["devicename", "iot_connected", "productkey", "wifi_connected"]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    IOT_CONNECTED_FIELD_NUMBER: _ClassVar[int]
    PRODUCTKEY_FIELD_NUMBER: _ClassVar[int]
    WIFI_CONNECTED_FIELD_NUMBER: _ClassVar[int]
    devicename: str
    iot_connected: bool
    productkey: str
    wifi_connected: bool
    def __init__(self, wifi_connected: bool = ..., iot_connected: bool = ..., productkey: _Optional[str] = ..., devicename: _Optional[str] = ...) -> None: ...

class mnet_inet_status(_message.Message):
    __slots__ = ["connect", "gateway", "ip", "mask"]
    CONNECT_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_FIELD_NUMBER: _ClassVar[int]
    IP_FIELD_NUMBER: _ClassVar[int]
    MASK_FIELD_NUMBER: _ClassVar[int]
    connect: bool
    gateway: int
    ip: int
    mask: int
    def __init__(self, connect: bool = ..., ip: _Optional[int] = ..., mask: _Optional[int] = ..., gateway: _Optional[int] = ...) -> None: ...

class WifiConfType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DrvUploadFileFileType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class DrvDevInfoResult(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class sim_card_sta(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class mnet_link_type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class apn_auth_type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class net_type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class iot_conctrl_type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from abc import ABC, abstractmethod
from typing import Any

from bleak import BLEDevice

from pymammotion import CloudIOTGateway
from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.model.device import MowingDevice, RTKDevice
from pymammotion.data.model.enums import ConnectionPreference
from pymammotion.mammotion.devices.mammotion_cloud import MammotionCloud


class AbstractDeviceManager(ABC):
    """Abstract base class for device managers."""

    def __init__(
        self,
        name: str,
        iot_id: str,
        cloud_client: CloudIOTGateway,
        cloud_device: Device,
        preference: ConnectionPreference = ConnectionPreference.BLUETOOTH,
    ) -> None:
        self.name = name
        self.iot_id = iot_id
        self.cloud_client = cloud_client
        self._device: Device = cloud_device
        self.mammotion_http = cloud_client.mammotion_http
        self.preference = preference

    @property
    @abstractmethod
    def state(self) -> MowingDevice | RTKDevice:
        """Return the state of the device."""

    @state.setter
    @abstractmethod
    def state(self, value: MowingDevice | RTKDevice) -> None:
        """Set the device state."""

    @property
    @abstractmethod
    def ble(self) -> Any | None:
        """Return BLE device interface."""

    @property
    @abstractmethod
    def cloud(self) -> Any | None:
        """Return cloud device interface."""

    @abstractmethod
    def has_queued_commands(self) -> bool:
        """Check if there are queued commands."""

    @abstractmethod
    def add_ble(self, ble_device: BLEDevice) -> Any:
        """Add BLE device."""

    @abstractmethod
    def add_cloud(self, mqtt: MammotionCloud) -> Any:
        """Add cloud device."""

    @abstractmethod
    def replace_cloud(self, cloud_device: Any) -> None:
        """Replace cloud device."""

    @abstractmethod
    def remove_cloud(self) -> None:
        """Remove cloud device."""

    @abstractmethod
    def replace_ble(self, ble_device: Any) -> None:
        """Replace BLE device."""

    @abstractmethod
    def remove_ble(self) -> None:
        """Remove BLE device."""

    @abstractmethod
    def replace_mqtt(self, mqtt: MammotionCloud) -> None:
        """Replace MQTT connection."""



from dataclasses import dataclass, field
from typing import Annotated, Generic, Literal, TypeVar

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Alias


class UnauthorizedException(Exception):
    pass


DataT = TypeVar("DataT")


@dataclass
class ErrorInfo(DataClassDictMixin):
    code: str
    platform: str
    module: str
    variant: str
    level: str
    description: str
    en_implication: str
    en_solution: str
    zh_implication: str
    zh_solution: str
    de_implication: str
    de_solution: str
    fr_implication: str
    fr_solution: str
    it_implication: str
    it_solution: str
    ja_implication: str
    ja_solution: str
    es_implication: str
    es_solution: str
    cs_implication: str
    cs_solution: str
    ko_implication: str
    ko_solution: str
    sk_implication: str
    sk_solution: str
    pl_implication: str
    pl_solution: str
    nl_implication: str
    nl_solution: str
    da_implication: str
    da_solution: str
    sr_implication: str
    sr_solution: str
    sv_implication: str
    sv_solution: str
    sl_implication: str
    sl_solution: str
    pt_implication: str
    pt_solution: str
    hu_implication: str
    hu_solution: str
    hr_implication: str
    hr_solution: str
    no_implication: str
    no_solution: str
    fi_implication: str
    fi_solution: str
    ro_implication: str
    ro_solution: str
    bg_implication: str
    bg_solution: str
    et_implication: str
    et_solution: str
    lv_implication: str
    lv_solution: str
    lt_implication: str
    lt_solution: str


@dataclass
class SettingVo(DataClassORJSONMixin):
    """Device setting configuration."""

    type: int = 0
    is_switch: Annotated[int, Alias("isSwitch")] = 0


@dataclass
class LocationVo(DataClassORJSONMixin):
    """Device location information."""

    date_time: Annotated[str, Alias("dateTime")] = ""
    date_timestamp: Annotated[int, Alias("dateTimestamp")] = 0
    location: list[float] = field(default_factory=lambda: [0.0, 0.0])


@dataclass
class DeviceInfo:
    """Complete device information."""

    iot_id: Annotated[str, Alias("iotId")] = ""
    device_id: Annotated[str, Alias("deviceId")] = ""
    device_name: Annotated[str, Alias("deviceName")] = ""
    device_type: Annotated[str, Alias("deviceType")] = ""
    series: str = ""
    product_series: Annotated[str, Alias("productSeries")] = ""
    icon_code: Annotated[str, Alias("iconCode")] = ""
    generation: int = 0
    status: int = 0
    is_subscribe: Annotated[int, Alias("isSubscribe")] = 0
    setting_vos: Annotated[list[SettingVo], Alias("settingVos")] = field(default_factory=list)
    active_time: Annotated[str, Alias("activeTime")] = ""
    active_timestamp: Annotated[int, Alias("activeTimestamp")] = 0
    location_vo: Annotated[LocationVo | None, Alias("locationVo")] = None

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True


@dataclass
class DeviceRecord(DataClassORJSONMixin):
    identity_id: Annotated[str, Alias("identityId")]
    iot_id: Annotated[str, Alias("iotId")]
    product_key: Annotated[str, Alias("productKey")]
    device_name: Annotated[str, Alias("deviceName")]
    owned: int
    status: int
    bind_time: Annotated[int, Alias("bindTime")]
    create_time: Annotated[str, Alias("createTime")]

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True


@dataclass
class DeviceRecords(DataClassORJSONMixin):
    records: list[DeviceRecord]
    total: int
    size: int
    current: int
    pages: int


@dataclass
class MQTTConnection(DataClassORJSONMixin):
    host: str
    jwt: str
    client_id: Annotated[str, Alias("clientId")]
    username: str


@dataclass
class Response(DataClassORJSONMixin, Generic[DataT]):
    code: int
    msg: str
    request_id: Annotated[str, Alias("requestId")] | None = None
    data: DataT | None = None

    class Config(BaseConfig):
        omit_default = True


@dataclass
class LoginResponseUserInformation(DataClassORJSONMixin):
    areaCode: str
    domainAbbreviation: str
    userId: str
    userAccount: str
    authType: str
    email: str | None = None

    class Config(BaseConfig):
        omit_none = True


@dataclass
class JWTTokenInfo(DataClassORJSONMixin):
    """specifically for newer devices and mqtt"""

    iot: str  # iot domain e.g api-iot-business-eu-dcdn.mammotion.com
    robot: str  # e.g api-robot-eu.mammotion.com


@dataclass
class LoginResponseData(DataClassORJSONMixin):
    access_token: str
    token_type: Literal["bearer", "Bearer"]
    refresh_token: str
    expires_in: int
    authorization_code: str
    userInformation: LoginResponseUserInformation
    jti: str = None
    grant_type: Literal["password", "Password"] = None
    scope: Literal["read", "Read"] = None

    class Config(BaseConfig):
        omit_none = True


@dataclass
class FirmwareVersions(DataClassORJSONMixin):
    firmware_version: Annotated[str, Alias("firmwareVersion")] = ""
    firmware_code: Annotated[str, Alias("firmwareCode")] = ""
    firmware_latest_version: Annotated[str, Alias("firmwareLatestVersion")] = ""
    firmware_type: Annotated[str, Alias("firmwareType")] = ""

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True


@dataclass
class ProductVersionInfo(DataClassORJSONMixin):
    release_note: Annotated[str, Alias("releaseNote")] = ""
    release_version: Annotated[str, Alias("releaseVersion")] = ""
    data_location: str | None = None

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True


@dataclass
class CheckDeviceVersion(DataClassORJSONMixin):
    cause_code: Annotated[int, Alias("causeCode")] = 0
    product_version_info_vo: Annotated[ProductVersionInfo | None, Alias("productVersionInfoVo")] = None
    progress: int | None = 0
    upgradeable: bool = False
    device_id: Annotated[str, Alias("deviceId")] = ""
    device_name: Annotated[str | None, Alias("deviceName")] = ""
    current_version: Annotated[str, Alias("currentVersion")] = ""
    isupgrading: bool | None = False
    cause_msg: Annotated[str, Alias("causeMsg")] = ""

    class Config(BaseConfig):
        allow_deserialization_not_by_alias = True

    def __eq__(self, other):
        if not isinstance(other, CheckDeviceVersion):
            return NotImplemented

        if self.device_id != other.device_id or self.current_version != other.current_version:
            return False

        if self.product_version_info_vo and other.product_version_info_vo:
            if self.product_version_info_vo.release_version != other.product_version_info_vo.release_version:
                return False
            return True
        elif self.product_version_info_vo is None and other.product_version_info_vo is None:
            return False
        else:
            return True
